import os.path
from datetime import datetime
import pytz
from pydrive2.drive import GoogleDrive

from .gdrive_utils import *
from .download import retrieve_relevant_ids

def get_datetime():
    tz = pytz.timezone('Europe/Moscow')
    now = datetime.now(tz)

    dt_string = now.strftime("%d-%m-%Y %H:%M:%S")
    return dt_string


def save_file_to_gdrive(data_router,
                        expname,
                        path_to_file,
                        link=None,
                        destination=None,
                        force_rewriting=False,
                        gauth=None):

    drive = GoogleDrive(gauth)

    if link is None:
        row = data_router[data_router['Эксперимент'] == expname]
        links = dict(zip(row.columns, row.values[0]))
        if destination not in links:
            raise ValueError(f'Wrong folder name: {destination}')
        link = links[destination]

    fid = id_from_link(link)

    dataname = os.path.basename(path_to_file)
    if force_rewriting:
        return_code, rel = retrieve_relevant_ids(link,
                                                 dataname,
                                                 whitelist=[],
                                                 extensions=['.npz', '.csv', 'xlsx'])
        if len(rel) != 0:
            if len(rel) > 1:
                print('More than one relevant file found, which is suspicious. Consider manual check')

            existing_file_data = rel[0]
            dat_id, dat_name = existing_file_data
            f = drive.CreateFile({'id': dat_id})

        else:
            print(f'Data for {dataname} not found in folder, nothing to rewrite')
            f = drive.CreateFile({'title': dataname, "parents": [{"kind": "drive#fileLink", "id": fid}]})

    else:
        date_time = get_datetime()
        ext = os.path.splitext(path_to_file)[1]
        dataname = dataname[:-len(ext)] + date_time + ext
        f = drive.CreateFile({'title': dataname, "parents": [{"kind": "drive#fileLink", "id": fid}]})

    f.SetContentFile(path_to_file)
    f.Upload()



from .dr_base import *
from .data import MVData
from .graph import ProximityGraph
from .embedding import Embedding


# TODO: refactor this
def dr_series(d,
              n_jumps,
              all_metric_params,
              all_graph_params,
              all_embedding_params,
              recalculate_if_error=0):

    print('---------------------------  JUMP 1  --------------------------------')
    print('Performing jump from dim', d.dim, 'to dim', all_embedding_params['dim'][0], ':')
    metric_params = dict(zip(all_metric_params.keys(), [val[0] for val in all_metric_params.values()]))
    graph_params = dict(zip(all_graph_params.keys(), [val[0] for val in all_graph_params.values()]))
    embedding_params = dict(zip(all_embedding_params.keys(), [val[0] for val in all_embedding_params.values()]))
    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]

    m_params = m_param_filter(metric_params)
    g_params = g_param_filter(graph_params)
    e_params = e_param_filter(embedding_params)

    maxiter = 20
    n_iter = maxiter * int(recalculate_if_error) + 1
    it = 0
    success = 0
    while it < n_iter and success == 0:
        try:
            emb = d.get_embedding(m_params, g_params, e_params)
            success = 1
        except:
            it += 1
            print("Unexpected error:", sys.exc_info()[0])
            raise

    if it == n_iter:
        raise Exception('First jump failed after %s attempts' % n_iter)

    if n_jumps > 1:
        datalist = [d, MVData(emb.coords, labels=d.labels)]

    for i in range(1, n_jumps):
        print('---------------------------  JUMP ' + str(i + 1) + ' --------------------------------', )
        print('Performing jump from dim', datalist[i].dim, 'to dim', all_embedding_params['dim'][i], ':')
        metric_params = dict(zip(all_metric_params.keys(), [val[i] for val in all_metric_params.values()]))
        graph_params = dict(zip(all_graph_params.keys(), [val[i] for val in all_graph_params.values()]))
        embedding_params = dict(zip(all_embedding_params.keys(), [val[i] for val in all_embedding_params.values()]))
        embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]

        m_params = m_param_filter(metric_params)
        g_params = g_param_filter(graph_params)
        e_params = e_param_filter(embedding_params)

        maxiter = 20
        n_iter = maxiter * (recalculate_if_error) + 1
        it = 0
        success = 0

        while it < n_iter and success == 0:
            try:
                emb = datalist[i].get_embedding(m_params, g_params, e_params)
                success = 1
            except:
                print('iter', it)
                it += 1

        if it == n_iter:
            raise Exception('Jump ', str(i + 1), ' failed after %s attempts' % n_iter)

        datalist.append(MVData(emb.coords, labels=d.labels))

    return emb



import hashlib

import scipy.sparse as ssp
from sklearn.preprocessing import MinMaxScaler
from scipy.signal import hilbert
import numpy as np
import scipy.stats as st
from numba import njit


def populate_nested_dict(content, outer, inner):
    nested_dict = {o: {} for o in outer}
    for o in outer:
        nested_dict[o] = {i: content.copy() for i in inner}

    return nested_dict


def add_names_to_nested_dict(datadict, names1, names2):
    # renaming for convenience
    n1 = len(datadict.keys())
    n2 = len(datadict[list(datadict.keys())[0]])

    if not (names1 is None and names2 is None):
        if names1 is None:
            names1 = range(n1)
        if names2 is None:
            names2 = range(n2)

        renamed_dict = populate_nested_dict(dict(), names1, names2)
        for i in range(n1):
            for j in range(n2):
                renamed_dict[names1[i]][names2[j]].update(datadict[i][j])
        return renamed_dict

    else:
        return datadict


def retrieve_relevant_from_nested_dict(nested_dict,
                                       target_key,
                                       target_value,
                                       operation='=',
                                       allow_missing_keys=False):
    relevant_pairs = []
    for key1 in nested_dict.keys():
        for key2 in nested_dict[key1].keys():
            data = nested_dict[key1][key2]
            if target_key not in data and not allow_missing_keys:
                raise ValueError(f'Target key {target_key} not found in data dict')

            if operation == '=':
                criterion = data.get(target_key) == target_value
            elif operation == '>':
                criterion = data.get(target_key) > target_value if data.get(target_key) is not None else False
            elif operation == '<':
                criterion = data.get(target_key) < target_value if data.get(target_key) is not None else False
            else:
                raise ValueError(f'Operation should be one of "=", ">", "<", but {operation} was passed')

            if criterion:
                relevant_pairs.append((key1, key2))

    return relevant_pairs


def rescale(data):
    scaler = MinMaxScaler(feature_range=(0, 1))
    res = scaler.fit_transform(data.reshape(-1, 1)).ravel()
    return res


def get_hash(data):
    # Prepare the object hash
    hash_id = hashlib.md5()
    hash_id.update(repr(data).encode('utf-8'))
    return hash_id.hexdigest()


def phase_synchrony(vec1, vec2):
    al1 = np.angle(hilbert(vec1), deg=False)
    al2 = np.angle(hilbert(vec2), deg=False)
    phase_sync = 1-np.sin(np.abs(al1-al2)/2)
    return phase_sync


def correlation_matrix_old(a, b):
    if np.allclose(a, b):
        return np.corrcoef(a,a)
    else:
        n1 = a.shape[0]
        n2 = b.shape[0]
        corrmat = np.zeros((n1, n2))
        for i in range(n1):
            for j in range(n2):
                corrmat[i,j] = st.pearsonr(a[i,:], b[j,:])[0]

        return corrmat


def correlation_matrix(A):
    '''
    # fast implementation.
    A: numpy array of shape (ndims, nvars)

    returns: numpy array of shape (nvars, nvars)
    '''

    am = A - np.mean(A, axis=1, keepdims=True)
    return am @ am.T / np.sum(am**2, axis=1, keepdims=True).T


def cross_correlation_matrix(A, B):
    '''
    # fast implementation.

    A: numpy array of shape (ndims, nvars1)
    B: numpy array of shape (ndims, nvars2)

    returns: numpy array of shape (nvars1, nvars2)
    '''
    am = A - np.mean(A, axis=0, keepdims=True)
    bm = B - np.mean(B, axis=0, keepdims=True)
    return am.T @ bm / (np.sqrt(np.sum(am**2, axis=0, keepdims=True)).T * np.sqrt(np.sum(bm**2, axis=0, keepdims=True)))


# TODO: review this function
def norm_cross_corr(a, b):
    a = (a - np.mean(a)) / (np.std(a) * len(a))
    b = (b - np.mean(b)) / (np.std(b))
    c = np.correlate(a, b, 'full')
    return c


def to_numpy_array(data):
    if isinstance(data, np.ndarray):
        return data

    if ssp.issparse(data):
        return data.A
    else:
        return np.array(data)




import numpy as np
import warnings
import tqdm
from itertools import combinations
import pickle

from ..signals.sig_base import TimeSeries
from .neuron import DEFAULT_MIN_BEHAVIOUR_TIME, Neuron
from .wavelet_event_detection import WVT_EVENT_DETECTION_PARAMS, extract_wvt_events, events_to_ts_array, ridges_to_containers
from ..utils.data import get_hash, populate_nested_dict
from ..information.info_base import get_1d_mi

STATS_VARS = ['data_hash', 'opt_delay', 'pre_pval', 'pre_rval', 'pval', 'rval', 'mi', 'rel_mi_beh', 'rel_mi_ca']
SIGNIFICANCE_VARS = ['stage1', 'shuffles1', 'stage2', 'shuffles2', 'final_p_thr', 'multicomp_corr', 'pairwise_pval_thr']
DEFAULT_STATS = dict(zip(STATS_VARS, [None for _ in STATS_VARS]))
DEFAULT_SIGNIFICANCE = dict(zip(SIGNIFICANCE_VARS, [None for _ in SIGNIFICANCE_VARS]))


def check_dynamic_features(dynamic_features):
    dfeat_lengths = {}
    for feat_id in dynamic_features:
        current_ts = dynamic_features[feat_id]
        if isinstance(current_ts, TimeSeries):
            len_ts = len(current_ts.data)
        else:
            len_ts = len(current_ts)

        dfeat_lengths[feat_id] = len_ts

    if len(set(dfeat_lengths.values())) != 1:
        print(dfeat_lengths)
        raise ValueError('Dynamic features have different lengths!')


class Experiment():
    '''
    Class for all Ca2+ experiment types

    Attributes
    ----------

    Methods
    -------

    '''

    def __init__(self, signature, calcium, spikes, exp_identificators,
                 static_features, dynamic_features, **kwargs):

        fit_individual_t_off = kwargs.get('fit_individual_t_off', False)
        reconstruct_spikes = kwargs.get('reconstruct_spikes', 'wavelet')
        bad_frames_mask = kwargs.get('bad_frames_mask', None)
        spike_kwargs = kwargs.get('spike_kwargs', None)

        check_dynamic_features(dynamic_features)
        self.exp_identificators = exp_identificators
        self.signature = signature

        for idx in exp_identificators:
            setattr(self, idx, exp_identificators[idx])

        if calcium is None:
            raise AttributeError('No calcium data provided')

        if reconstruct_spikes is None:
            if spikes is None:
                warnings.warn('No spike data provided, spikes reconstruction from Ca2+ data disabled')
        else:
            if spikes is not None:
                warnings.warn(f'Spike data will be overridden by reconstructed spikes from Ca2+ data with method={reconstruct_spikes}')

            if reconstruct_spikes == 'wavelet':
                print('Reconstructing events with wavelet method...')
                wvt_kwargs = WVT_EVENT_DETECTION_PARAMS.copy()
                wvt_kwargs['fps'] = static_features.get('fps')
                if spike_kwargs is not None:
                    for k, v in spike_kwargs.items():
                        wvt_kwargs[k] = v
                st_ev_inds, end_ev_inds, all_ridges = extract_wvt_events(calcium, wvt_kwargs)
                self._all_wvt_ridges = [ridges_to_containers(ridges) for ridges in all_ridges]
                spikes = events_to_ts_array(calcium.shape[1], st_ev_inds, end_ev_inds, wvt_kwargs['fps'])

            else:
                raise ValueError(f'"{reconstruct_spikes} method for event reconstruction is not available"')

        self.filtered_flag = False
        if bad_frames_mask is not None:
            calcium, spikes, dynamic_features = self._trim_data(calcium,
                                                                spikes,
                                                                dynamic_features,
                                                                bad_frames_mask)
        else:
            for feat_id in dynamic_features.copy():
                if not isinstance(dynamic_features[feat_id], TimeSeries):
                    dynamic_features[feat_id] = TimeSeries(dynamic_features[feat_id])

        self.n_cells = calcium.shape[0]
        self.n_frames = calcium.shape[1]

        self.neurons = []
        self.calcium = np.zeros(calcium.shape)
        if spikes is None:
            self.spikes = np.zeros(calcium.shape)
        else:
            self.spikes = spikes

        print('Building neurons...')
        for i in tqdm.tqdm(np.arange(self.n_cells), position=0, leave=True):
            cell = Neuron(str(i),
                          calcium[i, :],
                          self.spikes[i, :],
                          default_t_rise=static_features.get('t_rise_sec'),
                          default_t_off=static_features.get('t_off_sec'),
                          fps=static_features.get('fps'),
                          fit_individual_t_off=fit_individual_t_off)

                #cell.reconstruct_spikes(**kwargs)

            self.calcium[i, :] = cell.ca.data
            self.neurons.append(cell)

        self.dynamic_features = dynamic_features
        for feat_id in dynamic_features:
            setattr(self, feat_id, dynamic_features[feat_id])

        for sfeat_name in static_features:
            setattr(self, sfeat_name, static_features[sfeat_name])

        # attribute for a threshold pvalue for each pair p-value. Computed from desired FWER and
        # multiple hypothesis correction method
        self._pairwise_pval_thr = None

        #neuron-feature pair statistics
        self.stats_table = self._populate_cell_feat_dict(DEFAULT_STATS, fbunch=None, cbunch=None)

        #neuron-feature pair significance-related data
        self.significance_table = self._populate_cell_feat_dict(DEFAULT_SIGNIFICANCE, fbunch=None, cbunch=None)

        print('Building data hashes...')
        self._build_data_hashes(mode='calcium')
        if reconstruct_spikes is not None or spikes is not None:
            self._build_data_hashes(mode='spikes')

        print('Final checkpoint...')
        self._checkpoint()
        #self._load_precomputed_data(**kwargs)

        print(f'Experiment "{self.signature}" constructed successfully with {self.n_cells} neurons and {len(self.dynamic_features)} features')


    def check_ds(self, ds):
        if not hasattr(self, 'fps'):
            raise ValueError(f'fps not set for {self.signature}')

        time_step = 1.0/self.fps
        if time_step*ds>DEFAULT_MIN_BEHAVIOUR_TIME:
            print('Downsampling constant is too high: some behaviour acts may be skipped. '
                  f'Current minimal behaviour time interval is set to {DEFAULT_MIN_BEHAVIOUR_TIME} sec, '
                  f'downsampling {ds} will create time gaps of {time_step*ds} sec')

    def _build_pair_hash(self, cell_id, feat_id, mode='calcium'):
        '''
        Builds a unique hash-based representation of activity-feature pair data.
        feat_id should be a string or an iterable of strings (in case of joint MI calculation).
        '''
        if mode == 'calcium':
            act = self.neurons[cell_id].ca.data
        elif mode == 'spikes':
            act = self.neurons[cell_id].sp.data
        else:
            raise ValueError('"mode" can be either "calcium" or "spikes"')

        act_hash = get_hash(act)

        if (not isinstance(feat_id, str)) and len(feat_id) == 1:
            feat_id = feat_id[0]

        if isinstance(feat_id, str):
            dyn_data = self.dynamic_features[feat_id].data
            dyn_data_hash = get_hash(dyn_data)
            pair_hash = (act_hash, dyn_data_hash)

        else:
            ordered_fnames = tuple(sorted(list(feat_id)))
            list_of_hashes = [act_hash]
            for fname in ordered_fnames:
                dyn_data = self.dynamic_features[fname].data
                dyn_data_hash = get_hash(dyn_data)
                list_of_hashes.append(dyn_data_hash)

            pair_hash = tuple(list_of_hashes)

        return pair_hash

    def _build_data_hashes(self, mode='calcium'):
        '''
        Builds a unique hash-based representation of calcium-feature pair data for all cell-feature pairs..
        '''
        default_data_hashes = {dfeat: dict(zip(range(self.n_cells), [None for _ in range(self.n_cells)])) for dfeat in self.dynamic_features.keys()}
        self._data_hashes = {'calcium': default_data_hashes, 'spikes': default_data_hashes}
        for feat_id in self.dynamic_features:
            for cell_id in range(self.n_cells):
                self._data_hashes[mode][feat_id][cell_id] = self._build_pair_hash(cell_id, feat_id, mode=mode)

    def _trim_data(self, calcium, spikes, dynamic_features, bad_frames_mask, force_filter=False):

        if not force_filter and self.filtered_flag:
            raise AttributeError('Data is already filtered, if you want to force filtering it again, set "force_filter = True"')

        f_calcium = calcium[:, ~bad_frames_mask]
        if spikes is not None:
            f_spikes = spikes[:, ~bad_frames_mask]
        else:
            f_spikes = None

        f_dynamic_features = {}
        for feat_id in dynamic_features:
            current_ts = dynamic_features[feat_id]
            if isinstance(current_ts, TimeSeries):
                f_ts = TimeSeries(current_ts.data[~bad_frames_mask], discrete=current_ts.discrete)
            else:
                f_ts = TimeSeries(current_ts[~bad_frames_mask])

            f_dynamic_features[feat_id] = f_ts

        self.filtered_flag = True
        self.bad_frames_mask = bad_frames_mask

        return f_calcium, f_spikes, f_dynamic_features

    def _checkpoint(self):
        '''
        Check build for common errors
        '''
        if self.n_cells > self.n_frames:
            raise UserWarning('Number of cells > number of time frames, looks like the data is transposed')

        for dfeat in ['calcium', 'spikes']:
            if self.n_frames not in getattr(self, dfeat).shape:
                raise ValueError(f'"{dfeat}" feature has inappropriate shape: {getattr(self, dfeat).data.shape}'
                f'inconsistent with data length {self.n_frames}')

        for dfeat in self.dynamic_features.keys():
            if self.n_frames not in getattr(self, dfeat).data.shape:
                raise ValueError(f'"{dfeat}" feature has inappropriate shape: {getattr(self, dfeat).data.shape}'
                 f'inconsistent with data length {self.n_frames}')

    def _populate_cell_feat_dict(self, content, fbunch=None, cbunch=None):
        '''
        Helper function. Creates a nested dictionary of feature-cell pairs and populates every cell with 'content' variable.
        Outer dict: dynamic features, inner dict: cells
        '''
        cell_ids = self._process_cbunch(cbunch)
        feat_ids = self._process_fbunch(fbunch, allow_multifeatures=True)
        nested_dict = populate_nested_dict(content, feat_ids, cell_ids)
        return nested_dict

    def _process_cbunch(self, cbunch):
        '''
        Helper function. Turns cell indices (int, iterable or None) into a list of cell numbers.
        '''
        if isinstance(cbunch, int):
            cell_ids = [cbunch]
        elif cbunch is None:
            cell_ids = list(np.arange(self.n_cells))
        else:
            cell_ids = list(cbunch)

        return cell_ids

    def _process_fbunch(self, fbunch, allow_multifeatures=False):
        '''
        Helper function. Turns feature names (str, iterable or None) into a list of feature names
        '''
        if isinstance(fbunch, str):
            feat_ids = [fbunch]

        elif fbunch is None:  # default set of features
            if allow_multifeatures:
                try:
                    # stats table contains up-to-date set of features, including multifeatures
                    feat_ids = list(self.stats_table.keys())
                except:
                    # if stats is not available, take full set of single features
                    feat_ids = list(self.dynamic_features.keys())
            else:
                feat_ids = list(self.dynamic_features.keys())

        else:
            feat_ids = []

            # check for multifeatures
            for fname in fbunch:
                if isinstance(fname, str):
                    feat_ids.append(fname)
                else:
                    if allow_multifeatures:
                        feat_ids.append(tuple(sorted(list(fname))))
                    else:
                        raise ValueError('Multifeature detected in "allow_multifeatures=False" mode')

        return feat_ids

    def _process_sbunch(self, sbunch, significance_mode=False):
        '''
        Helper function. Turns stats type names (str, iterable or None) into a list of stats types
        '''
        if significance_mode:
            default_list = SIGNIFICANCE_VARS
        else:
            default_list = STATS_VARS

        if isinstance(sbunch, str):
            return [sbunch]

        elif sbunch is None:
            return default_list

        else:
            return [st for st in sbunch if st in default_list]

    def _add_multifeature_to_data_hashes(self, feat_id, mode='calcium'):
        '''
        Add previously unseen multifeature (e.g. ['x','y']) to table with data hashes.
        This function ignores multifeatures that already exist in the table.
        '''
        if (not isinstance(feat_id, str)) and len(feat_id) == 1:
            feat_id = feat_id[0]

        if not isinstance(feat_id, str):
            ordered_fnames = tuple(sorted(list(feat_id)))
            if ordered_fnames not in self._data_hashes[mode]:
                all_hashes = [self._build_pair_hash(cell_id, ordered_fnames) for cell_id in range(self.n_cells)]
                new_dict = {ordered_fnames: dict(zip(range(self.n_cells), all_hashes))}
                self._data_hashes[mode].update(new_dict)

        else:
            raise ValueError('This method is for multifeature update only')

    def _add_multifeature_to_stats(self, feat_id):
        '''
        Add previously unseen multifeature (e.g. ['x','y']) to statistics and significance tables.
        This function ignores multifeatures that already exist in the table.
        '''
        if (not isinstance(feat_id, str)) and len(feat_id) == 1:
            feat_id = feat_id[0]

        if not isinstance(feat_id, str):
            ordered_fnames = tuple(sorted(list(feat_id)))
            if ordered_fnames not in self.stats_table:
                print(f'Multifeature {feat_id} is new, it will be added to stats table')
                self.stats_table[ordered_fnames] = {cell_id: DEFAULT_STATS.copy() for cell_id in range(self.n_cells)}
                self.significance_table[ordered_fnames] = {cell_id: DEFAULT_SIGNIFICANCE.copy() for cell_id in range(self.n_cells)}

        else:
            raise ValueError('This method is for multifeature update only')

    def _check_stats_relevance(self, cell_id, feat_id, mode='calcium'):
        '''
        A guardian function that prevents access to non-existing and irrelevant data.

        This function checks whether the calcium-feature pair statistics has already been calculated.
        It ensures the data (both calcium and dynamic feature) has not changed since the last
        calculation by checking hash values of both data arrays.

        This function always refers to stats table but works equally well with significance table
        since they are always updated simultaneously
        '''

        if not isinstance(feat_id, str):
            feat_id = tuple(sorted(list(feat_id)))

        if feat_id not in self.stats_table:
            raise ValueError(f'Feature {feat_id} is not present in stats. \n If this is a single feature, '
                             'check the input data, since all single features are processed automatically.'
                             'If this is a multifeature (e.g. ["x", "y"]), compute MI significance to create stats')

        pair_hash = self._data_hashes[mode][feat_id][cell_id]
        existing_hash = self.stats_table[feat_id][cell_id]['data_hash']

        # if (stats does not exist yet) or (stats exists and data is the same):
        if (existing_hash is None) or (pair_hash == existing_hash):
            return True

        else:
            print(f'Looks like the data for the pair (cell {cell_id}, feature {feat_id}) '
                  'has been changed since the last calculation)')

            return False

    def _update_stats_and_significance(self, stats, cell_id, feat_id, stage2_only):
        '''
        Updates stats table and linked significance table to erase irrelevant data properly
        '''
        # update statistics
        self.stats_table[feat_id][cell_id].update(stats)
        if not stage2_only:
            # erase significance data completely since stats for stage 1 has been modified
            self.significance_table[feat_id][cell_id].update(DEFAULT_SIGNIFICANCE.copy())
        else:
            # erase significance data for stage 2 since stats for stage 2 has been modified
            self.significance_table[feat_id][cell_id].update({'stage2': None, 'shuffles2': None})

    def update_neuron_feature_pair_stats(self, stats, cell_id, feat_id, force_update=False, stage2_only=False):
        '''
        Updates calcium-feature pair statistics.
        feat_id should be a string or an iterable of strings (in case of joint MI calculation).
        This function allows multifeatures.
        '''

        if not isinstance(feat_id, str):
            self._add_multifeature_to_data_hashes(feat_id)
            self._add_multifeature_to_stats(feat_id)

        if self._check_stats_relevance(cell_id, feat_id):
            self._update_stats_and_significance(stats, cell_id, feat_id, stage2_only=stage2_only)

        else:
            if not force_update:
                print(f'To forcefully update the stats, set "force_update=True"')
            else:
                self._update_stats_and_significance(stats, cell_id, feat_id, stage2_only=stage2_only)

    def update_neuron_feature_pair_significance(self, sig, cell_id, feat_id, mode='calcium'):
        '''
        Updates calcium-feature pair significance data.
        feat_id should be a string or an iterable of strings (in case of joint MI calculation).
        This function allows multifeatures.
        '''
        if not isinstance(feat_id, str):
            self._add_multifeature_to_data_hashes(feat_id, mode=mode)
            self._add_multifeature_to_stats(feat_id)

        if self._check_stats_relevance(cell_id, feat_id):
            self.significance_table[feat_id][cell_id].update(sig)

        else:
            raise ValueError('Can not update significance table until the collision between actual data hashes and '
                             'saved stats data hashes is resolved. Use update_neuron_feature_pair_stats' 
                             'with "force_update=True" to forcefully rewrite statistics')

    def get_neuron_feature_pair_stats(self, cell_id, feat_id):
        '''
        Returns calcium-feature pair statistics.
        This function allows multifeatures.
        '''
        stats = None
        if self._check_stats_relevance(cell_id, feat_id):
            stats = self.stats_table[feat_id][cell_id]
        else:
            print(f'Consider recalculating stats')

        return stats

    def get_neuron_feature_pair_significance(self, cell_id, feat_id):
        '''
        Returns calcium-feature pair significance data.
        This function allows multifeatures.
        '''
        sig = None
        if self._check_stats_relevance(cell_id, feat_id):
            sig = self.significance_table[feat_id][cell_id]
        else:
            print(f'Consider recalculating stats')

        return sig

    def get_multicell_shuffled_calcium(self, cbunch=None, method='roll_based', no_ts=True, **kwargs):
        '''

        Args:
            cbunch:
            method:
            **kwargs:
            no_ts: if True, intermediate TimeSeries objects are nor created, which speeds up shuffling

        Returns:

        '''
        cell_list = self._process_cbunch(cbunch)
        agg_sh_data = np.zeros((len(cell_list), self.n_frames))
        for i in cell_list:
            cell = self.neurons[i]
            sh_data = cell.get_shuffled_calcium(method=method, **kwargs, no_ts=no_ts)
            if no_ts:
                agg_sh_data[i, :] = sh_data[:]
            else:
                agg_sh_data[i, :] = sh_data.data[:]

        return agg_sh_data

    def get_multicell_shuffled_spikes(self, cbunch = None, method = 'isi_based', no_ts=True, **kwargs):
        '''

        Args:
            cbunch:
            method:
            **kwargs:
            no_ts: if True, intermediate TimeSeries objects are nor created, which speeds up shuffling

        Returns:

        '''
        if self.spikes is None:
            raise AttributeError('Unable to shuffle spikes without spikes data')

        cell_list = self._process_cbunch(cbunch)

        agg_sh_data = np.zeros((len(cell_list), self.n_frames))
        for i in cell_list:
            cell = self.neurons[i]
            sh_data = cell.get_shuffled_spikes(method=method, **kwargs)
            if no_ts:
                agg_sh_data[i, :] = sh_data[:]
            else:
                agg_sh_data[i, :] = sh_data.data[:]

        return agg_sh_data

    def get_stats_slice(self, table_to_scan=None, cbunch=None, fbunch=None, sbunch=None, significance_mode=False):
        '''
        returns slice of accumulated statistics data (or significance data if "significance_mode=True")
        '''
        cell_ids = self._process_cbunch(cbunch)
        feat_ids = self._process_fbunch(fbunch, allow_multifeatures=True)
        slist = self._process_sbunch(sbunch, significance_mode=significance_mode)

        if table_to_scan is None:
            if significance_mode:
                full_table = self.significance_table
            else:
                full_table = self.stats_table
        else:
            full_table = table_to_scan

        out_table = self._populate_cell_feat_dict(dict(), fbunch=fbunch, cbunch=cbunch)
        for feat_id in feat_ids:
            for cell_id in cell_ids:
                out_table[feat_id][cell_id] = {s: full_table[feat_id][cell_id][s] for s in slist}

        return out_table

    def get_significance_slice(self, cbunch=None, fbunch=None, sbunch=None):
        return self.get_stats_slice(cbunch=cbunch, fbunch=fbunch, sbunch=sbunch, significance_mode=True)

    def get_feature_entropy(self, feat_id, ds=1):
        '''
        Calculates entropy of a single dynamic feature or a multifeature (e.g. ['x','y']).
        Currently only 2-combinations of features are correctly supported,
        for 3 and more variables calculations will be distorted (correct estimation of multivariate
        entropy for non-gaussian variables is non-trivial).
        '''
        if isinstance(feat_id, str):
            fts = self.dynamic_features[feat_id]
            return fts.get_entropy(ds=ds)

        else:
            ordered_fnames = tuple(sorted(list(feat_id)))
            tslist = [self.dynamic_features[dfeat] for dfeat in ordered_fnames]
            single_entropies = [fts.get_entropy(ds=ds) for fts in tslist]
            fpairs = list(combinations(tslist, 2))
            MIs = [get_1d_mi(ts1, ts2, ds=ds) for (ts1,ts2) in fpairs]
            return sum(single_entropies) - sum(MIs)

    def get_significant_neurons(self, min_nspec=1, fbunch=None):
        '''
        Returns a dict with neuron ids as keys and their significantly correlated features as values
        Only neurons with "min_nspec" or more significantly correlated features will be returned
        '''
        feat_ids = self._process_fbunch(fbunch, allow_multifeatures=True)
        relevance = [self._check_stats_relevance(cell_id, feat_id) for cell_id in range(self.n_cells) for feat_id in feat_ids]
        if not np.all(np.array(relevance)):
            raise ValueError('Stats relevance error')

        cell_ids = np.arange(self.n_cells)

        # TODO: add significance update and pval_thr argument
        cell_feat_dict = {cell_id: [] for cell_id in range(self.n_cells)}
        for i, cell_id in enumerate(cell_ids):
            for j, feat_id in enumerate(feat_ids):
                if self.significance_table[feat_id][cell_id]['stage2']:
                    cell_feat_dict[cell_id].append(feat_id)

        # filter out cells without specializations
        final_cell_feat_dict = {cell_id: cell_feat_dict[cell_id] for cell_id in range(self.n_cells) if len(cell_feat_dict[cell_id])>=min_nspec}

        return final_cell_feat_dict


    #===================================================================================
    # not active

    def save_mi_significance_to_file(self, fname):
        with open(fname, 'wb') as f:
            pickle.dump(self.mi_significance_table, f)


    def clear_cells_mi_significance_data(self, cbunch, path_to_save = None):
        for cell_id in cbunch:
            for feat in self.dynamic_features:
                self.mi_significance_table[feat][cell_id] = {}

        if path_to_save is not None:
            self.save_mi_significance_to_file(path_to_save)

    def clear_features_mi_significance_data(self, feat_list, save_to_file = False):
        pass

    def clear_cell_feat_mi_significance_data(self, cell, feat, save_to_file = False):
        pass

    def _load_precomputed_data(self, **kwargs):
        if 'mi_significance' in kwargs:
            self.mi_significance_table = {**self.mi_significance_table, **kwargs['mi_significance']}



def construct_session_name(data_source, exp_params, allow_unknown=True):
    if data_source == 'IABS':
        track = exp_params['track']
        animal_id, session = exp_params['animal_id'], exp_params['session']

        if track == 'HT':
            name = f'{animal_id}_HT{session}'
        elif track == 'RT':
            name = f'RT_{animal_id}_{session}D'
        elif track == 'FS':
            name = f'FS{animal_id}_{session}D'
        elif track == 'FcOY':
            name = f'FcOY_{animal_id}_{session}'
        elif track == 'STFP':
            name = f'STFP_{animal_id}_{session}'
        elif track == 'AP':
            name = f'AP_{animal_id}_{session}'
        elif track == 'NOF':
            name = f'NOF_{animal_id}_{session}'
        elif track == 'Trace':
            name = f'Trace_{animal_id}_{session}'
        elif track == 'CC':
            name = f'CC_{animal_id}_{session}'
        else:
            if not allow_unknown:
                raise ValueError(f'Unknown track: {track}!')
            else:
                name = f'{track}_{animal_id}_{session}'

        return name

    else:
        raise NotImplementedError('Other data sources are not yet supported')


import networkx as nx
import numpy as np
from copy import deepcopy

from .randomization import *

def get_giant_cc_from_graph(G):
    # this function preserves graph type: nx.Graph --> nx.Graph; nx.DiGraph --> nx.DiGraph
    connected_components = sorted(nx.connected_components(G), key=len, reverse=True)
    # print([len(c) for c in connected_components])
    gcc = connected_components[0]

    return nx.subgraph(G, gcc)


def get_giant_scc_from_graph(G):
    # for a directed graph, its largest strongly connected component is returned.
    if nx.is_directed(G):
        strongly_connected_components = sorted(nx.strongly_connected_components(G),
                                               key=len,
                                               reverse=True)

        gcc = strongly_connected_components[0]

    else:
        raise ValueError('Strongly connected components are meaningless for undirected graphs')

    return nx.subgraph(G, gcc)


def remove_selfloops_from_graph(graph):
    g = deepcopy(graph) # NetworkX graphs are highly nested, deepcopy is safer
    # this function preserves graph type: nx.Graph --> nx.Graph; nx.DiGraph --> nx.DiGraph
    g.remove_edges_from(list(nx.selfloop_edges(g)))
    return g


def remove_isolates_and_selfloops_from_graph(graph):
    g = deepcopy(graph) # NetworkX graphs are highly nested, deepcopy is safer
    # this function preserves graph type: nx.Graph --> nx.Graph; nx.DiGraph --> nx.DiGraph
    g.remove_nodes_from(list(nx.isolates(g)))
    g.remove_edges_from(list(nx.selfloop_edges(g)))
    return g


def remove_isolates_from_graph(graph):
    g = deepcopy(graph)  # NetworkX graphs are highly nested, deepcopy is safer
    g.remove_nodes_from(list(nx.isolates(g)))
    return g


def small_world_index(G, nrand=10, null_model='erdos-renyi'):
    asp = nx.average_shortest_path_length(G)
    acc = nx.average_clustering(G)

    r_asp_list = []
    r_acc_list = []
    i = 0
    while i < nrand:
        if null_model == 'maslov-sneppen':
            #ar = adj_random_rewiring_iom_preserving(sp.csr_array(data), is_weighted=True).A
            raise Exception('not implemented yet')

        elif null_model == 'erdos-renyi':
            n = G.number_of_nodes()
            p = 2.0 * G.number_of_edges() / n / (n - 1)
            Gr = nx.gnp_random_graph(n, p)

        if nx.is_connected(Gr):
            i += 1
            r_asp_list.append(nx.average_shortest_path_length(Gr))
            r_acc_list.append(nx.average_clustering(Gr))

    sw = (acc / np.mean(r_acc_list)) / (asp / np.mean(r_asp_list))
    return sw



import tqdm
import imageio
import matplotlib.pyplot as plt
import numpy as np
from os import listdir, remove
from os.path import isfile, join, splitext

from .plot import *


def erase_all(path, signature='', ext='.png'):
    # destroy all previous images
    prev_files_paths = [join(path, f) for f in listdir(path) if signature in f]
    files_to_del = [fp for fp in prev_files_paths if isfile(fp) and splitext(fp) == ext]
    for fp in files_to_del:
        remove(fp)


def save_image_series(path, figures, im_ext='png'):
    #with io.capture_output() as captured:

    for i in tqdm.tqdm(np.arange(1, len(figures)), leave = True, position = 0):
        fig = figures[i]
        figname = fig._suptitle.get_text() + im_ext
        fig.savefig(join(path, figname))
        fig.close()


def create_gif_from_image_series(path, signature, gifname, erase_prev=True, im_ext='png', duration=0.2):
    if erase_prev:
        erase_all(path, signature=signature, ext=im_ext)

    images = []
    imfiles = [f for f in listdir(path) if isfile(join(path, f)) and signature in f and im_ext in f]
    imfiles = sorted(imfiles)

    for filename in tqdm.tqdm(imfiles, leave=True, position=0):
        images.append(imageio.v3.imread((join(path, filename))))

    imageio.mimsave(join(path, 'GIFs', f'{signature} {gifname}.gif'), images, duration=duration)



import numpy as np
import numba as nb
from numba import njit
from scipy.special import ndtri, psi, digamma

from .info_utils import py_fast_digamma

#TODO: credits to original GCMI

def ctransform(x):
    """Copula transformation (empirical CDF)
    cx = ctransform(x) returns the empirical CDF value along the first
    axis of x. Data is ranked and scaled within [0 1] (open interval).
    """

    xi = np.argsort(np.atleast_2d(x))
    xr = np.argsort(xi)
    cx = (xr + 1).astype(float) / (xr.shape[-1] + 1)
    return cx


def copnorm(x):
    """Copula normalization

    cx = copnorm(x) returns standard normal samples with the same empirical
    CDF value as the input. Operates along the last axis.
    """
    # cx = sp.stats.norm.ppf(ctransform(x))
    cx = ndtri(ctransform(x))
    return cx


@njit()
def demean(x):
    return x - np.full(len(x), x.mean())


@njit()
def ent_g(x, biascorrect=True):
    """Entropy of a Gaussian variable in bits
    H = ent_g(x) returns the entropy of a (possibly
    multidimensional) Gaussian variable x with bias correction.
    Columns of x correspond to samples, rows to dimensions/variables.
    (Samples last axis)
    """
    x = np.atleast_2d(x)
    if x.ndim > 2:
        raise ValueError("x must be at most 2d")
    Ntrl = x.shape[1]
    Nvarx = x.shape[0]

    # demean data
    x = demean(x)
    # covariance
    C = np.dot(x, x.T) / float(Ntrl - 1)
    chC = np.linalg.cholesky(C)

    # entropy in nats
    HX = np.sum(np.log(np.diagonal(chC))) + 0.5 * Nvarx * (np.log(2 * np.pi) + 1.0)

    ln2 = np.log(2)
    if biascorrect:
        psiterms = py_fast_digamma((Ntrl - np.arange(1, Nvarx + 1).astype(float)) / 2.0) / 2.0
        dterm = (ln2 - np.log(Ntrl - 1.0)) / 2.0
        HX = HX - Nvarx * dterm - psiterms.sum()

    # convert to bits
    return HX / ln2


@njit()
def mi_gg(x, y, biascorrect=True, demeaned=False):
    """Mutual information (MI) between two Gaussian variables in bits

    I = mi_gg(x,y) returns the MI between two (possibly multidimensional)
    Gassian variables, x and y, with bias correction.
    If x and/or y are multivariate columns must correspond to samples, rows
    to dimensions/variables. (Samples last axis)

    biascorrect : true / false option (default true) which specifies whether
    bias correction should be applied to the esimtated MI.
    demeaned : false / true option (default false) which specifies whether th
    input data already has zero mean (true if it has been copula-normalized)
    """

    x = np.atleast_2d(x)
    y = np.atleast_2d(y)
    if x.ndim > 2 or y.ndim > 2:
        raise ValueError("x and y must be at most 2d")
    Ntrl = x.shape[1]
    Nvarx = x.shape[0]
    Nvary = y.shape[0]
    Nvarxy = Nvarx + Nvary

    if y.shape[1] != Ntrl:
        raise ValueError("number of trials do not match")

    # joint variable
    xy = np.vstack((x, y))

    if not demeaned:
        xy = demean(xy)

    Cxy = np.dot(xy, xy.T) / float(Ntrl - 1)
    # submatrices of joint covariance
    Cx = Cxy[:Nvarx, :Nvarx]
    Cy = Cxy[Nvarx:, Nvarx:]

    chCxy = np.linalg.cholesky(Cxy)
    chCx = np.linalg.cholesky(Cx)
    chCy = np.linalg.cholesky(Cy)

    # entropies in nats
    # normalizations cancel for mutual information
    HX = np.sum(np.log(np.diag(chCx)))  # + 0.5*Nvarx*(np.log(2*np.pi)+1.0)
    HY = np.sum(np.log(np.diag(chCy)))  # + 0.5*Nvary*(np.log(2*np.pi)+1.0)
    HXY = np.sum(np.log(np.diag(chCxy)))  # + 0.5*Nvarxy*(np.log(2*np.pi)+1.0)

    ln2 = np.log(2)
    if biascorrect:
        psiterms = py_fast_digamma((Ntrl - np.arange(1, Nvarxy + 1)) / 2.0) / 2.0
        dterm = (ln2 - np.log(Ntrl - 1.0)) / 2.0
        HX = HX - Nvarx * dterm - psiterms[:Nvarx].sum()
        HY = HY - Nvary * dterm - psiterms[:Nvary].sum()
        HXY = HXY - Nvarxy * dterm - psiterms[:Nvarxy].sum()

    # MI in bits
    I = (HX + HY - HXY) / ln2
    return I


@njit()
def mi_model_gd(x, y, Ym, biascorrect=True, demeaned=False):
    """Mutual information (MI) between a Gaussian and a discrete variable in bits
    based on ANOVA style model comparison.
    I = mi_model_gd(x,y,Ym) returns the MI between the (possibly multidimensional)
    Gaussian variable x and the discrete variable y.
    For 1D x this is a lower bound to the mutual information.
    Columns of x correspond to samples, rows to dimensions/variables.
    (Samples last axis)
    y should contain integer values in the range [0 Ym-1] (inclusive).
    biascorrect : true / false option (default true) which specifies whether
    bias correction should be applied to the estimated MI.
    demeaned : false / true option (default false) which specifies whether the
    input data already has zero mean (true if it has been copula-normalized)
    See also: mi_mixture_gd
    """

    x = np.atleast_2d(x)
    # y = np.squeeze(y)
    if x.ndim > 2:
        raise ValueError("x must be at most 2d")
    if y.ndim > 1:
        raise ValueError("only univariate discrete variables supported")
    '''
    if not np.issubdtype(y.dtype, np.integer):
        raise ValueError("y should be an integer array")
    '''
    if int(Ym) != Ym:
        raise ValueError("Ym should be an integer")

    Ntrl = x.shape[1]
    Nvarx = x.shape[0]

    if y.size != Ntrl:
        raise ValueError("number of trials do not match")
    '''
    if not demeaned:
        x = x - x.mean(axis=1)[:,np.newaxis]
    '''
    # class-conditional entropies
    Ntrl_y = np.zeros(Ym)
    Hcond = np.zeros(Ym)
    c = 0.5 * (np.log(2.0 * np.pi) + 1)

    for yi in range(Ym):
        idx = y == yi
        xm = x[:, idx]
        Ntrl_y[yi] = xm.shape[1]
        xm = demean(xm)
        Cm = np.dot(xm, xm.T) / float(Ntrl_y[yi] - 1)
        chCm = np.linalg.cholesky(Cm)
        Hcond[yi] = np.sum(np.log(np.diag(chCm)))  # + c*Nvarx

    # class weights
    w = Ntrl_y / float(Ntrl)

    # unconditional entropy from unconditional Gaussian fit
    Cx = np.dot(x, x.T) / float(Ntrl - 1)
    chC = np.linalg.cholesky(Cx)
    Hunc = np.sum(np.log(np.diag(chC)))  # + c*Nvarx

    ln2 = np.log(2)
    if biascorrect:
        vars = np.arange(1, Nvarx + 1)

        psiterms = py_fast_digamma((Ntrl - vars) / 2.0) / 2.0
        dterm = (ln2 - np.log(float(Ntrl - 1))) / 2.0
        Hunc = Hunc - Nvarx * dterm - psiterms.sum()

        dterm = (ln2 - np.log((Ntrl_y - 1))) / 2.0
        psiterms = np.zeros(Ym)
        for vi in vars:
            idx = Ntrl_y - vi
            psiterms = psiterms + py_fast_digamma(idx / 2.0)
        Hcond = Hcond - Nvarx * dterm - (psiterms / 2.0)

    # MI in bits
    I = (Hunc - np.sum(w * Hcond)) / ln2
    return I


def gcmi_cc(x, y):
    """Gaussian-Copula Mutual Information between two continuous variables.
    I = gcmi_cc(x,y) returns the MI between two (possibly multidimensional)
    continuous variables, x and y, estimated via a Gaussian copula.
    If x and/or y are multivariate columns must correspond to samples, rows
    to dimensions/variables. (Samples first axis)
    This provides a lower bound to the true MI value.
    """

    x = np.atleast_2d(x)
    y = np.atleast_2d(y)
    if x.ndim > 2 or y.ndim > 2:
        raise ValueError("x and y must be at most 2d")
    Ntrl = x.shape[1]
    Nvarx = x.shape[0]
    Nvary = y.shape[0]

    if y.shape[1] != Ntrl:
        raise ValueError("number of trials do not match")

    '''
    # check for repeated values
    for xi in range(Nvarx):
        if (np.unique(x[xi,:]).size / float(Ntrl)) < 0.9:
            warnings.warn("Input x has more than 10% repeated values")
            break
    for yi in range(Nvary):
        if (np.unique(y[yi,:]).size / float(Ntrl)) < 0.9:
            warnings.warn("Input y has more than 10% repeated values")
            break
    '''

    # copula normalization
    cx = copnorm(x)
    cy = copnorm(y)
    # parametric Gaussian MI
    I = mi_gg(cx, cy, True, True)
    return I



import matplotlib.pyplot as plt
import matplotlib.pylab as pylab


def make_beautiful(ax):
    for axis in ['bottom', 'left']:
        ax.spines[axis].set_linewidth(4)
        ax.tick_params(width=4, direction='in', length=8, pad=15)

    for axis in ['top', 'right']:
        ax.spines[axis].set_linewidth(0.0)

    # ax.locator_params(axis='x', nbins=8)
    # ax.locator_params(axis='y', nbins=8)
    ax.tick_params(axis='x', which='major', labelsize=26)
    ax.tick_params(axis='y', which='major', labelsize=26)

    ax.xaxis.label.set_size(30)
    ax.yaxis.label.set_size(30)

    params = {'legend.fontsize': 18,
              'axes.titlesize': 30,
              }

    pylab.rcParams.update(params)

    return ax


def create_default_figure(a=16, b=12):
    fig, ax = plt.subplots(figsize=(a, b))
    ax = make_beautiful(ax)

    return fig, ax


def plot_mat(mat, a=12, b=12, ax=None, with_cbar=True):
    if ax is None:
        fig, ax = plt.subplots(figsize=(a, b))

    im = ax.imshow(mat)
    if with_cbar:
        cbar = ax.figure.colorbar(im, ax=ax)

    return ax



import numpy as np
from numba import njit

@njit()
def py_fast_digamma(data):
    res = np.zeros(len(data))
    for i, x in enumerate(data):
        "Faster digamma function assumes x > 0."
        r = 0
        while x <= 5:
            r -= 1 / x
            x += 1
        f = 1 / (x * x)
        t = f * (-1 / 12.0 + f * (1 / 120.0 + f * (-1 / 252.0 + f * (1 / 240.0 + f * (-1 / 132.0
                                                                                      + f * (691 / 32760.0 + f * (-1 / 12.0 + f * 3617 / 8160.0)))))))

        res[i] = r + np.log(x) - 0.5 / x + t

    return res


def binary_mi_score(contingency):
    nzx, nzy = np.nonzero(contingency)
    nz_val = contingency[nzx, nzy]

    contingency_sum = contingency.sum()
    pi = np.ravel(contingency.sum(axis=1))
    pj = np.ravel(contingency.sum(axis=0))

    # Since MI <= min(H(X), H(Y)), any labelling with zero entropy, i.e. containing a
    # single cluster, implies MI = 0
    if pi.size == 1 or pj.size == 1:
        return 0.0

    log_contingency_nm = np.log(nz_val)
    contingency_nm = nz_val / contingency_sum
    # Don't need to calculate the full outer product, just for non-zeroes
    outer = pi.take(nzx).astype(np.int64, copy=False) * pj.take(nzy).astype(
        np.int64, copy=False
    )
    log_outer = -np.log(outer) + np.log(pi.sum()) + np.log(pj.sum())
    mi = (
            contingency_nm * (log_contingency_nm - np.log(contingency_sum))
            + contingency_nm * log_outer
    )
    mi = np.where(np.abs(mi) < np.finfo(mi.dtype).eps, 0.0, mi)
    return np.clip(mi.sum(), 0.0, None)


import numpy as np

import matplotlib.pyplot as plt
from ..utils.plot import create_default_figure, make_beautiful
from ..utils.data import rescale
from scipy.stats import rankdata, gaussian_kde, wasserstein_distance
import seaborn as sns


def plot_pc_activity(exp, cell_ind, ds=None, ax=None):
    pc_stats = exp.stats_table[('x', 'y')][cell_ind]
    pval = None if pc_stats['pval'] is None else np.round(pc_stats['pval'], 7)
    rel_mi_beh = None if pc_stats['rel_mi_beh'] is None else np.round(pc_stats['rel_mi_beh'], 4)

    if ds is None:
        ds = 5

    if ax is None:
        lenx = max(exp.x.data) - min(exp.x.data)
        leny = max(exp.y.data) - min(exp.y.data)
        xyratio = max(lenx / leny, leny / lenx)
        fig, ax = create_default_figure(6*xyratio, 6)

    #neur = np.roll(rescale(rankdata(exp.neurons[ind].ca.data)), 0)
    neur = rescale(np.log(exp.neurons[cell_ind].ca.data+1e-10))
    spinds = np.where(exp.neurons[cell_ind].sp.data != 0)[0]

    #ax.plot(exp.x.data[::ds], exp.y.data[::ds], c = 'k', alpha=0.3)
    ax.scatter(exp.x.data[::ds], exp.y.data[::ds], c=neur[::ds], cmap = 'plasma', alpha=0.8)
    ax.scatter(exp.x.data[spinds], exp.y.data[spinds], c='k', alpha=1, marker='*', linewidth=2, s=100)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_title(f'Cell {cell_ind}, Rel MI={rel_mi_beh}, pval={pval}')

    return ax


def plot_neuron_feature_density(exp, data_type, cell_id, featname, ind1=0, ind2=100000, ds=1, shift=None, ax=None):
    ind2 = min(exp.n_frames, ind2)

    if data_type == 'calcium':
        sig = exp.neurons[cell_id].ca.scdata[ind1:ind2][::ds]
    if data_type == 'spikes':
        sig = exp.neurons[cell_id].sp.scdata[ind1:ind2][::ds]

    feature = getattr(exp, featname)
    bdata = feature.scdata[ind1:ind2][::ds]
    rbdata = rescale(rankdata(bdata))

    if ax is None:
        fig, ax = plt.subplots(figsize=(6,6))

    if feature.is_binary:
        if data_type == 'calcium':
            vals0 = np.log10(sig[np.where((rbdata == min(rbdata)) & (sig > 0))])
            vals1 = np.log10(sig[np.where((rbdata == max(rbdata)) & (sig > 0))])

            wsd = wasserstein_distance(vals0, vals1)
            # _ = ax.hist(vals0, bins = 25, color = 'b', log = True, density = True, alpha = 0.7, label=f'{featname}=1')
            _ = sns.kdeplot(vals0, ax=ax, c='b', label=f'{featname}=0', linewidth=3, bw_adjust=0.5)
            _ = sns.kdeplot(vals1, ax=ax, c='r', label=f'{featname}=1', linewidth=3, bw_adjust=0.5)
            # _ = ax.hist(vals1, bins = 25, color = 'r', log = True, density = True, alpha = 0.7, label=f'{featname}=0')
            ax.legend(loc='upper right')
            ax.set_xlabel('log(dF/F)', fontsize=20)
            ax.set_ylabel('density', fontsize=20)
            ax.set_title(f'wsd={wsd}')

        if data_type == 'spikes':
            raise Exception('Not implemented yet')

    #TODO: finish this function
    else:
        x0, y0 = np.log10(sig + np.random.random(size=len(sig)) * 1e-8), np.log(
            bdata + np.random.random(size=len(bdata)) * 1e-8)

        jdata = np.vstack([x0, y0]).T
        # jplot = sns.jointplot(jdata, x=jdata[:,0], y=jdata[:,1], kind='hist', bins=100)
        nbins = 100
        k = gaussian_kde(jdata.T)
        xi, yi = np.mgrid[x0.min():x0.max():nbins * 1j, y0.min():y0.max():nbins * 1j]
        zi = k(np.vstack([xi.flatten(), yi.flatten()]))

        # plot a density
        ax.set_title('Density')
        ax.pcolormesh(xi, yi, zi.reshape(xi.shape), shading='auto', cmap='coolwarm')
        ax.set_xlabel('log(signals)', fontsize=20)
        ax.set_ylabel(f'log({featname})', fontsize=20)

    return ax


def plot_neuron_feature_pair(exp, cell_id, featname, ind1=0, ind2=100000, ds=1,
                             add_density_plot=True, ax=None, title=None):

    ind2 = min(exp.n_frames, ind2)
    ca = exp.neurons[cell_id].ca.scdata[ind1:ind2][::ds]
    #rca = rescale(rankdata(ca))
    feature = getattr(exp, featname)
    bdata = feature.scdata[ind1:ind2][::ds]
    rbdata = rescale(rankdata(bdata))

    if ax is None:
        if add_density_plot:
            fig, axs = plt.subplots(1, 2, figsize=(12, 6), width_ratios=[0.6, 0.4])
            ax0, ax1 = axs
            ax1 = make_beautiful(ax1)
        else:
            fig, ax0 = plt.subplots(figsize=(10, 6), width_ratios=[0.6, 0.4])
            ax1 = None

    ax0 = make_beautiful(ax0)

    ax0.plot(np.arange(ind1, ind2)[::ds], ca, c='b', linewidth=2, alpha=0.5, label=f'neuron {cell_id}')
    if feature.discrete:
        ax0.scatter(np.arange(ind1, ind2)[::ds][np.where(rbdata == 1)], ca[np.where(rbdata == 1)], c='r', linewidth=2)
    else:
        ax0.plot(np.arange(ind1, ind2)[::ds], rbdata, c='r', linewidth=2, alpha=0.5)

    if add_density_plot:
        plot_neuron_feature_density(exp, cell_id, featname, ind1=ind1, ind2=ind2, ds=ds, ax=ax1)

    ax0.set_xlabel('timeframes', fontsize=20)
    ax0.set_ylabel('Signal/behavior', fontsize=20)

    if title is None:
        title = f'{exp.signature} Neuron {cell_id}, feature {featname}'

    fig.suptitle(title, fontsize=20)
    plt.tight_layout()

    return fig



import numpy as np
from scipy.stats import gamma, lognorm, rankdata
from ..utils.data import populate_nested_dict, add_names_to_nested_dict
from ..experiment.exp_base import DEFAULT_STATS

def chebyshev_ineq(data, val):
    z = (val - np.mean(data))/np.std(data)
    return 1./z**2


def get_lognormal_p(data, val):
    params = lognorm.fit(data, floc=0)
    rv = lognorm(*params)
    return rv.sf(val)


def get_gamma_p(data, val):
    params = gamma.fit(data, floc=0)
    rv = lognorm(*params)
    return rv.sf(val)


def get_mi_distr_pvalue(data, val, distr_type='gamma'):
    if distr_type == 'lognormal':
        distr = lognorm
    elif distr_type == 'gamma':
        distr = gamma
    else:
        raise ValueError(f'wrong MI distribution: {distr_type}')

    params = distr.fit(data, floc=0)
    rv = distr(*params)
    return rv.sf(val)


def get_mask(ptable, rtable, pval_thr, rank_thr):
    mask = np.ones(ptable.shape)
    mask[np.where(ptable > pval_thr)] = 0
    mask[np.where(rtable < rank_thr)] = 0
    return mask


def stats_not_empty(pair_stats, current_data_hash, stage=1):
    if stage == 1:
        stats_to_check = ['pre_rval', 'pre_pval']
    elif stage == 2:
        stats_to_check = ['rval', 'pval', 'mi']
    else:
        raise ValueError(f'Stage should be 1 or 2, but {stage} was passed')

    data_hash_from_stats = pair_stats['data_hash']
    is_valid = (current_data_hash == data_hash_from_stats)
    is_not_empty = np.all(np.array([pair_stats[st] is not None for st in stats_to_check]))
    return is_valid and is_not_empty


def criterion1(pair_stats, nsh1, topk=1):
    """
    Calculates whether the given neuron-feature pair is potentially significant after preliminary shuffling

    Parameters
    ----------
    pair_stats: dict
        dictionary of computed stats

    nsh1: int
        number of shuffles for first stage

    topk: int
        true MI should be among topk MI shuffles
        default: 1

    Returns
    -------
    crit_passed: bool
        True if significance confirmed, False if not.
    """

    if pair_stats.get('pre_rval') is not None:
        return pair_stats['pre_rval'] > (1 - 1.*topk/(nsh1+1))
        #return pair_stats['pre_rval'] == 1 # true MI should be top-1 among all shuffles
    else:
        return False


def criterion2(pair_stats, nsh2, pval_thr, topk=5):
    """
    Calculates whether the given neuron-feature pair is significant after full-scale shuffling

    Parameters
    ----------
    pair_stats: dict
        dictionary of computed stats

    nsh2: int
        number of shuffles for second stage

    pval_thr: float
        pvalue threshold for a single pair. It depends on a FWER significance level and multiple
        hypothesis correction algorithm.

    topk: int
        true MI should be among topk MI shuffles
        default: 5

    Returns
    -------
    crit_passed: bool
        True if significance confirmed, False if not.
    """
    # whether pair passed stage 1 and has statistics from stage 2
    if pair_stats.get('rval') is not None and pair_stats.get('pval') is not None:
        # whether true MI is among topk shuffles (in practice it is top-1 almost always)
        if pair_stats['rval'] > (1 - 1.*topk/(nsh2+1)):
            criterion = pair_stats['pval'] < pval_thr
            return criterion
        else:
            return False
    else:
        return False


def get_all_nonempty_pvals(all_stats, ids1, ids2):
    all_pvals = []
    for i, id1 in enumerate(ids1):
        for j, id2 in enumerate(ids2):
            pval = all_stats[id1][id2].get('pval')
            if not(pval is None):
                all_pvals.append(pval)

    return all_pvals


def get_table_of_stats(mitable,
                       optimal_delays,
                       precomputed_mask=None,
                       mi_distr_type='gamma',
                       nsh=0,
                       stage=1):

    # 0 in mask values means that stats for this pair will not be calculated
    # 1 in mask values means that stats for this pair will be calculated from new results.
    if precomputed_mask is None:
        precomputed_mask = np.ones(mitable.shape)

    a, b, sh = mitable.shape
    stage_stats = populate_nested_dict(dict(), range(a), range(b))

    ranked_total_mi = rankdata(mitable, axis=2, nan_policy='omit')
    ranks = (ranked_total_mi[:, :, 0] / (nsh + 1))  # how many shuffles have MI lower than true mi

    for i in range(a):
        for j in range(b):
            if precomputed_mask[i, j]:
                new_stats = DEFAULT_STATS.copy()
                mi = mitable[i, j, 0]
                random_mi_samples = mitable[i, j, 1:]
                pval = get_mi_distr_pvalue(random_mi_samples, mi, distr_type=mi_distr_type)
                opt_delay = optimal_delays[i, j]

                if stage == 1:
                    new_stats['pre_rval'] = ranks[i, j]
                    new_stats['pre_pval'] = pval
                    new_stats['opt_delay'] = opt_delay

                elif stage == 2:
                    new_stats['rval'] = ranks[i,j]
                    new_stats['pval'] = pval
                    new_stats['mi'] = mitable[i,j,0]
                    new_stats['opt_delay'] = opt_delay

                stage_stats[i][j].update(new_stats)

    return stage_stats


def merge_stage_stats(stage1_stats, stage2_stats):
    merged_stats = stage2_stats.copy()
    for i in stage2_stats:
        for j in stage2_stats[i]:
            merged_stats[i][j]['pre_rval'] = stage1_stats[i][j]['pre_rval']
            merged_stats[i][j]['pre_pval'] = stage1_stats[i][j]['pre_pval']

    return merged_stats


def merge_stage_significance(stage_1_significance, stage_2_significance):
    merged_significance = stage_2_significance.copy()
    for i in stage_2_significance:
        for j in stage_2_significance[i]:
            merged_significance[i][j].update(stage_1_significance[i][j])

    return merged_significance



from scipy.stats import norm
import numpy as np

def brownian(x0, n, dt=1, delta=1, out=None):
    """
    Generate an instance of Brownian motion (i.e. the Wiener process):

        X(t) = X(0) + N(0, delta**2 * t; 0, t)

    where N(a,b; t0, t1) is a normally distributed random variable with mean a and
    variance b.  The parameters t0 and t1 make explicit the statistical
    independence of N on different time intervals; that is, if [t0, t1) and
    [t2, t3) are disjoint intervals, then N(a, b; t0, t1) and N(a, b; t2, t3)
    are independent.

    Written as an iteration scheme,

        X(t + dt) = X(t) + N(0, delta**2 * dt; t, t+dt)


    If `x0` is an array (or array-like), each value in `x0` is treated as
    an initial condition, and the value returned is a numpy array with one
    more dimension than `x0`.

    Arguments
    ---------
    x0 : float or numpy array (or something that can be converted to a numpy array
         using numpy.asarray(x0)).
        The initial condition(s) (i.e. position(s)) of the Brownian motion.
    n : int
        The number of steps to take.
    dt : float
        The time step.
    delta : float
        delta determines the "speed" of the Brownian motion.  The random variable
        of the position at time t, X(t), has a normal distribution whose mean is
        the position at time t=0 and whose variance is delta**2*t.
    out : numpy array or None
        If `out` is not None, it specifies the array in which to put the
        result.  If `out` is None, a new numpy array is created and returned.

    Returns
    -------
    A numpy array of floats with shape `x0.shape + (n,)`.

    Note that the initial value `x0` is not included in the returned array.
    """

    x0 = np.asarray(x0)

    # For each element of x0, generate a sample of n numbers from a
    # normal distribution.
    r = norm.rvs(size=x0.shape + (n,), scale=delta*np.sqrt(dt))

    # If `out` was not given, create an output array.
    if out is None:
        out = np.empty(r.shape)

    # This computes the Brownian motion by forming the cumulative sum of
    # the random samples.
    np.cumsum(r, axis=-1, out=out)

    # Add the initial condition.
    out += np.expand_dims(x0, axis=-1)

    return out


# Credit to Greg Ver Steeg (http://www.isi.edu/~gregv/npeet.html)

import numpy as np
import numpy.linalg as la
from numpy import log
from sklearn.neighbors import BallTree, KDTree

from .info_utils import py_fast_digamma

DEFAULT_NN = 5
# UTILITY FUNCTIONS

def add_noise(x, intens=1e-10):
    # small noise to break degeneracy, see doc.
    return x + intens * np.random.random_sample(x.shape)


def query_neighbors(tree, x, k):
    # return tree.query(x, k=k+1, breadth_first = False)[0][:, k]
    return tree.query(x, k=k + 1)[0][:, k]


def _count_neighbors_single(tree, x, radii, ind):
    dists, indices = tree.query(x[ind:ind + 1], k=DEFAULT_NN, distance_upper_bound=radii[ind])
    return len(np.unique(indices[0])) - 2


def count_neighbors(tree, x, radii):
    return tree.query_radius(x, radii, count_only=True)
    # dists, indices = tree.query(x, k=DEFAULT_NN, distance_upper_bound=r)
    # out = tree.query(x, k=DEFAULT_NN, distance_upper_bound=r)
    # return np.array([_count_neighbors_single(tree, x, radii, ind) for ind in range(len(x))])
    # return np.array([len(nn)-1 for nn in tree.query_ball_point(x, radii)])


def build_tree(points, lf=5):
    if points.shape[1] >= 20:
        return BallTree(points, metric='chebyshev')

    return KDTree(points, metric='chebyshev', leaf_size=lf)
    # return KDTree(points, leafsize = lf)
    # return KDTree(points, copy_data=True, leafsize = 5)


def avgdigamma(points, dvec, lf=30, tree=None):
    # This part finds number of neighbors in some radius in the marginal space
    # returns expectation value of <psi(nx)>
    if tree is None:
        tree = build_tree(points, lf=lf)

    dvec = dvec - 1e-15
    num_points = count_neighbors(tree, points, dvec)
    num_points = num_points.astype(float)

    zero_inds = np.where(num_points == 0)[0]
    if 1.0 * len(zero_inds) / len(num_points) > 0.01:
        raise Exception('No neighbours in more than 1% points, check input!')
    else:
        if len(zero_inds) != 0:
            num_points[zero_inds] = 0.5

    # inf_inds = np.where(digamma(num_points) == -np.inf)
    # print(num_points[inf_inds])

    digammas = list(map(py_fast_digamma, num_points))
    return np.mean(digammas)


# CONTINUOUS ESTIMATORS

def nonparam_entropy_c(x, k=DEFAULT_NN, base=np.e):
    """ The classic K-L k-nearest neighbor continuous entropy estimator.
    """
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    # xs_columns = np.expand_dims(xs, axis=0).T
    x = np.asarray(x)
    if len(x.shape) == 1:
        x = x.reshape(-1, 1)
    n_elements, n_features = x.shape
    x = add_noise(x)
    tree = build_tree(x)
    nn = query_neighbors(tree, x, k)
    const = py_fast_digamma(n_elements) - py_fast_digamma(k) + n_features * log(2)
    return (const + n_features * np.log(nn).mean()) / log(base)


def nonparam_cond_entropy_cc(x, y, k=DEFAULT_NN, base=np.e):
    """ The classic K-L k-nearest neighbor continuous entropy estimator for the
        entropy of X conditioned on Y.
    """
    xy = np.c_[x, y]
    entropy_union_xy = nonparam_entropy_c(xy, k=k, base=base)
    entropy_y = nonparam_entropy_c(y, k=k, base=base)
    return entropy_union_xy - entropy_y


def nonparam_mi_cc_mod(x, y, z=None, k=DEFAULT_NN, base=np.e, alpha=0,
                       lf=5, precomputed_tree_x=None, precomputed_tree_y=None):
    """
    Mutual information of x and y (conditioned on z if z is not None)
    """

    assert len(x) == len(y), "Arrays should have same length"
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"

    x, y = np.asarray(x), np.asarray(y)
    x, y = x.reshape(x.shape[0], -1), y.reshape(y.shape[0], -1)
    x = add_noise(x)
    y = add_noise(y)

    points = [x, y]
    if z is not None:
        z = np.asarray(z)
        z = z.reshape(z.shape[0], -1)
        points.append(z)

    points = np.hstack(points)

    # Find nearest neighbors in joint space, p=inf means max-norm
    tree = build_tree(points, lf=lf)
    dvec = query_neighbors(tree, points, k)

    if z is None:
        a = avgdigamma(x, dvec, tree=precomputed_tree_x, lf=lf)
        b = avgdigamma(y, dvec, tree=precomputed_tree_y, lf=lf)
        c = py_fast_digamma(k)
        d = py_fast_digamma(len(x))

        # print(a, b, c, d)

        if alpha > 0:
            d += lnc_correction(tree, points, k, alpha)
    else:
        xz = np.c_[x, z]
        yz = np.c_[y, z]
        a, b, c, d = avgdigamma(xz, dvec), avgdigamma(
            yz, dvec), avgdigamma(z, dvec), py_fast_digamma(k)

    return (-a - b + c + d) / log(base)


def lnc_correction(tree, points, k, alpha):
    e = 0
    n_sample = points.shape[0]
    for point in points:
        # Find k-nearest neighbors in joint space, p=inf means max norm
        knn = tree.query(point[None, :], k=k + 1, return_distance=False)[0]
        knn_points = points[knn]
        # Substract mean of k-nearest neighbor points
        knn_points = knn_points - knn_points[0]
        # Calculate covariance matrix of k-nearest neighbor points, obtain eigen vectors
        covr = knn_points.T @ knn_points / k
        _, v = la.eig(covr)
        # Calculate PCA-bounding box using eigen vectors
        V_rect = np.log(np.abs(knn_points @ v).max(axis=0)).sum()
        # Calculate the volume of original box
        log_knn_dist = np.log(np.abs(knn_points).max(axis=0)).sum()

        # Perform local non-uniformity checking and update correction term
        if V_rect < log_knn_dist + np.log(alpha):
            e += (log_knn_dist - V_rect) / n_sample
    return e


import cvxpy as cp
import numpy as np
from scipy.sparse.csgraph import laplacian
from sklearn.neighbors import NearestNeighbors

class MaximumVarianceUnfolding(object):

    def __init__(self, equation="berkley", solver=cp.SCS, solver_tol=1e-2,
                 eig_tol=1.0e-10, solver_iters=2500, warm_start=False, seed=None):
        """
        :param equation: A string either "berkley" or "wikipedia" to represent
                         two different equations for the same problem.
        :param solver: A CVXPY solver object.
        :param solver_tol: A float representing the tolerance the solver uses to know when to stop.
        :param eig_tol: The positive semi-definite constraint is only so accurate, this sets
                        eigenvalues that lie in -eig_tol < 0 < eig_tol to 0.
        :param solver_iters: The max number of iterations the solver will go through.
        :param warm_start: Whether or not to use a warm start for the solver.
                           Useful if you are running multiple tests on the same data.
        :param seed: The numpy seed for random numbers.
        """
        self.equation = equation
        self.solver = solver
        self.solver_tol = solver_tol
        self.eig_tol = eig_tol
        self.solver_iters = solver_iters
        self.warm_start = warm_start
        self.seed = seed
        self.neighborhood_graph = None

    def fit(self, data, k):
        """
        The method to fit an MVU model to the data.
        :param data: The data to which the model will be fitted.
        :param k: The number of neighbors to fix.
        :param dropout_rate: The number of neighbors to discount.
        :return: Embedded Gramian: The Gramian matrix of the embedded data.
        """
        # Number of data points in the set
        n = data.shape[0]

        # Set the seed
        np.random.seed(self.seed)

        # Calculate the nearest neighbors of each data point and build a graph
        N = NearestNeighbors(n_neighbors=k).fit(data).kneighbors_graph(data).todense()
        N = np.array(N)

        # Save the neighborhood graph to be accessed latter
        self.neighborhood_graph = N

        # To check for disconnected regions in the neighbor graph
        lap = laplacian(N, normed=True)
        eigvals, _ = np.linalg.eig(lap)

        for e in eigvals:
            if e == 0. and self.solver_iters is None:
                raise ValueError("DISCONNECTED REGIONS IN NEIGHBORHOOD GRAPH. "
                                      "PLEASE SPECIFY MAX ITERATIONS FOR THE SOLVER")

        # Declare some CVXPy variables
        # Gramian of the original data
        P = cp.Constant(data.dot(data.T))
        # The projection of the Gramian
        Q = cp.Variable((n, n), PSD=True)
        # Initialized to zeros
        Q.value = np.zeros((n, n))
        # A shorter way to call a vector of 1's
        ONES = cp.Constant(np.ones((n, 1)))
        # A variable to keep the notation consistent with the Berkley lecture
        T = cp.Constant(n)

        # Declare placeholders to get rid of annoying warnings
        objective = None
        constraints = []

        # Wikipedia Solution
        if self.equation == "wikipedia":
            objective = cp.Maximize(cp.trace(Q))

            constraints = [Q >> 0, cp.sum(Q, axis=1) == 0]

            for i in range(n):
                for j in range(n):
                    if N[i, j] == 1:
                        constraints.append((P[i, i] + P[j, j] - P[i, j] - P[j, i]) -
                                           (Q[i, i] + Q[j, j] - Q[i, j] - Q[j, i]) == 0)

        # UC Berkley Solution
        if self.equation == "berkley":
            objective = cp.Maximize(cp.multiply((1 / T), cp.trace(Q)) -
                                    cp.multiply((1 / (T * T)), cp.trace(cp.matmul(cp.matmul(Q, ONES), ONES.T))))

            constraints = [Q >> 0, cp.sum(Q, axis=1) == 0]
            for i in range(n):
                for j in range(n):
                    if N[i, j] == 1.:
                        constraints.append(Q[i, i] - 2 * Q[i, j] + Q[j, j] -
                                           (P[i, i] - 2 * P[i, j] + P[j, j]) == 0)

        # Solve the problem with the SCS Solver
        problem = cp.Problem(objective, constraints)
        # FIXME The solvertol syntax is unique to SCS
        problem.solve(solver=self.solver,
                      eps=self.solver_tol,
                      max_iters=self.solver_iters,
                      warm_start=self.warm_start)

        return Q.value


    def fit_transform(self, data, dim, k):
        """
        The method to fit and transform an MVU model to the data.
        :param data: The data to which the model will be fitted.
        :param dim: The new dimension of the dataset.
        :param k: The number of neighbors to fix.
        :param dropout_rate: The number of neighbors to discount.
        :return: embedded_data: The embedded form of the data.
        """

        embedded_gramian = self.fit(data, k)

        # Retrieve Q
        embedded_gramian = embedded_gramian

        # Decompose gramian to recover the projection
        eigenvalues, eigenvectors = np.linalg.eig(embedded_gramian)

        # Set the eigenvalues that are within +/- eig_tol to 0
        eigenvalues[np.logical_and(-self.eig_tol < eigenvalues, eigenvalues < self.eig_tol)] = 0.

        # Assuming the eigenvalues and eigenvectors aren't sorted,
        #    sort them and get the top "dim" ones
        sorted_indices = eigenvalues.argsort()[::-1]
        top_eigenvalue_indices = sorted_indices[:dim]

        # Take the top eigenvalues and eigenvectors
        top_eigenvalues = eigenvalues[top_eigenvalue_indices]
        top_eigenvectors = eigenvectors[:, top_eigenvalue_indices]

        # Some quick math to get the projection and return it
        lbda = np.diag(top_eigenvalues ** 0.5)
        embedded_data = lbda.dot(top_eigenvectors.T).T

        return embedded_data


from sklearn.feature_selection import mutual_info_classif, mutual_info_regression
from sklearn.metrics.cluster import mutual_info_score

from .ksg import *
from .gcmi import *
from .info_utils import binary_mi_score

import numpy as np
import warnings
from sklearn.preprocessing import MinMaxScaler
from scipy.stats import entropy, differential_entropy

from ..utils.data import to_numpy_array


DEFAULT_NN = 5

# TODO: add @property decorators to properly set getter-setter functionality

class TimeSeries():
    @staticmethod
    def define_ts_type(ts):
        if len(ts) < 100:
            warnings.warn('Time series is too short for accurate type (discrete/continuous) determination')

        unique_vals = np.unique(ts)
        sc1 = len(unique_vals) / len(ts)
        hist = np.histogram(ts, bins=len(ts))[0]
        ent = entropy(hist)
        maxent = entropy(np.ones(len(ts)))
        sc2 = ent / maxent

        # TODO: refactor thresholds
        if sc1 > 0.70 and sc2 > 0.70:
            return False  # both scores are high - the variable is most probably continuous
        elif sc1 < 0.25 and sc2 < 0.25:
            return True  # both scores are low - the variable is most probably discrete
        else:
            raise ValueError(f'Unable to determine time series type automatically: score 1 = {sc1}, score 2 = {sc2}')

    def _check_input(self):
        pass

    def __init__(self, data, discrete=None, shuffle_mask=None):
        self.data = to_numpy_array(data)

        if discrete is None:
            #warnings.warn('Time series type not specified and will be inferred automatically')
            self.discrete = TimeSeries.define_ts_type(self.data)
        else:
            self.discrete = discrete

        scaler = MinMaxScaler()
        self.scdata = scaler.fit_transform(self.data.reshape(-1, 1)).reshape(1, -1)[0]
        self.copula_normal_data = None

        if self.discrete:
            self.int_data = np.round(self.data).astype(int)
            if len(set(self.data.astype(int))) == 2:
                self.is_binary = True
                self.bool_data = self.int_data.astype(bool)
            else:
                self.is_binary = False

        else:
            self.copula_normal_data = copnorm(self.data).ravel()

        self.entropy = dict()  # supports various downsampling constants
        self.kdtree = None
        self.kdtree_query = None

        if shuffle_mask is None:
            # which shuffles are valid
            self.shuffle_mask = np.ones(len(self.data)).astype(bool)
        else:
            self.shuffle_mask = shuffle_mask.astype(bool)

    def get_kdtree(self):
        if self.kdtree is None:
            tree = self._compute_kdtree()
            self.kdtree = tree

        return self.kdtree

    def _compute_kdtree(self):
        d = self.scdata.reshape(self.scdata.shape[0], -1)
        return build_tree(d)

    def get_kdtree_query(self, k=DEFAULT_NN):
        if self.kdtree_query is None:
            q = self._compute_kdtree_query(k=k)
            self.kdtree_query = q

        return self.kdtree_query

    def _compute_kdtree_query(self, k=DEFAULT_NN):
        tree = self.get_kdtree()
        return tree.query(self.scdata, k=k + 1)

    def get_entropy(self, ds=1):
        if ds not in self.entropy.keys():
            self._compute_entropy(ds=ds)
        return self.entropy[ds]

    def _compute_entropy(self, ds=1):
        if self.discrete:
            counts = []
            for val in np.unique(self.data[::ds]):
                counts.append(len(np.where(self.data[::ds] == val)[0]))

            self.entropy[ds] = entropy(counts, base=np.e)

        else:
            # TODO: refactor entropy calculation
            self.entropy[ds] = get_tdmi(self.scdata[::ds], min_shift=1, max_shift=2)[0]
            #raise AttributeError('Entropy for continuous variables is not yet implemented'


def get_1d_mi(ts1, ts2, shift=0, ds=1, k=DEFAULT_NN, estimator='gcmi'):
    """Computes mutual information between two 1d variables efficiently

    Parameters
    ----------
    ts1: TimeSeries instance or numpy array
    ts2: TimeSeries instance or numpy array
    shift: int
        ts2 will be roll-moved by the number 'shift' after downsampling by 'ds' factor
    ds: int
        downsampling constant (take every 'ds'-th point)
    k: int
        number of neighbors for ksg estimator
    estimator: str
        Estimation method. Should be 'ksg' (accurate but slow) and 'gcmi' (ultra-fast, but estimates the lower bound on MI).
        In most cases 'gcmi' should be preferred.

    Returns
    -------
    mi: mutual information (or its lower bound in case of 'gcmi' estimator) between ts1 and (possibly) shifted ts2

    """
    if not isinstance(ts1, TimeSeries):
        ts1 = TimeSeries(ts1)
    if not isinstance(ts2, TimeSeries):
        ts2 = TimeSeries(ts2)

    if estimator == 'ksg':
        x = ts1.scdata[::ds].reshape(-1, 1)
        y = ts2.scdata[::ds]
        if shift != 0:
            y = np.roll(y, shift)

        if not ts1.discrete and not ts2.discrete:
            mi = nonparam_mi_cc_mod(ts1.scdata, y, k=k,
                                    precomputed_tree_x=ts1.get_kdtree(),
                                    precomputed_tree_y=ts2.get_kdtree())

        elif ts1.discrete and ts2.discrete:
            mi = mutual_info_classif(ts1.data, ts2.data, discrete_features=True, n_neighbors=k)[0]

        # TODO: refactor using ksg functions
        elif ts1.discrete and not ts2.discrete:
            mi = mutual_info_regression(ts1.data, y, discrete_features=False, n_neighbors=k)[0]

        elif not ts1.discrete and ts2.discrete:
            mi = mutual_info_classif(x, ts2.data, discrete_features=True, n_neighbors=k)[0]

        return mi

    elif estimator == 'gcmi':
        if not ts1.discrete and not ts2.discrete:
            ny1 = ts1.copula_normal_data[::ds]
            ny2 = np.roll(ts2.copula_normal_data[::ds], shift)
            mi = mi_gg(ny1, ny2, True, True)

        elif ts1.discrete and ts2.discrete:
            # if features are binary:
            if ts1.is_binary and ts2.is_binary:
                ny1 = ts1.bool_data[::ds]
                ny2 = np.roll(ts2.bool_data[::ds], shift)

                contingency = np.zeros((2, 2))
                contingency[0, 0] = (ny1 & ny2).sum()
                contingency[0, 1] = (~ny1 & ny2).sum()
                contingency[0, 1] = (ny1 & ~ny2).sum()
                contingency[1, 1] = (~ny1 & ~ny2).sum()

                mi = binary_mi_score(contingency)

            else:
                ny1 = ts1.int_data[::ds]  # .reshape(-1, 1)
                ny2 = np.roll(ts2.int_data[::ds], shift)
                mi = mutual_info_score(ny1, ny2)

        elif ts1.discrete and not ts2.discrete:
            ny1 = ts1.int_data[::ds]
            ny2 = np.roll(ts2.copula_normal_data[::ds], shift)
            mi = mi_model_gd(ny2, ny1, np.max(ny1), biascorrect=True, demeaned=True)

        elif not ts1.discrete and ts2.discrete:
            ny1 = ts1.copula_normal_data[::ds]
            #TODO: fix zd error
            ny2 = np.roll(ts2.int_data[::ds], shift)
            #ny2 = np.roll(ts2.data[::ds], shift)
            '''
            print(ny2)
            print(sum(ny2))
            print(ny1)
            '''

            mi = mi_model_gd(ny1, ny2, np.max(ny2), biascorrect=True, demeaned=True)

        if mi < 0:
            mi = 0

        return mi


def get_tdmi(data, min_shift=1, max_shift=100, nn=DEFAULT_NN):
    ts = TimeSeries(data, discrete=False)
    tdmi = [get_1d_mi(ts, ts, shift=shift, k=nn) for shift in range(min_shift, max_shift)]

    return tdmi


def get_multi_mi(tslist, ts2, shift=0, ds=1, k=DEFAULT_NN, estimator='gcmi'):

    #TODO: make shift the same as in get_1d_mi
    if ~np.all([ts.discrete for ts in tslist]) and not ts2.discrete:
        nylist = [ts.copula_normal_data[::ds] for ts in tslist]
        ny1 = np.vstack(nylist)
        ny2 = np.roll(ts2.copula_normal_data, shift)[::ds]
        mi = mi_gg(ny1, ny2, True, True)
    else:
        raise ValueError('Multidimensional MI only implemented for continuous data!')

    if mi < 0:
        mi = 0

    return mi


from ..utils.plot import *
from .matrix_utils import *
import numpy as np
import networkx as nx
from itertools import combinations
from matplotlib import cm
import matplotlib as mpl
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize as color_normalize


def draw_degree_distr(net, mode=None, cumulative=0, survival=1, log_log=0):
    if not net.directed:
        mode = 'all'

    fig, ax = create_default_figure(10, 8)
    ax.set_title('Degree distribution', color='white')

    if mode is not None:
        distr = net.get_degree_distr(mode=mode)
        if cumulative:
            if survival:
                distr = 1 - np.cumsum(distr)
            else:
                distr = np.cumsum(distr)

        if log_log:
            degree, = ax.loglog(distr[:-1], linewidth=2, c='k', label='degree')
        else:
            degree, = ax.plot(distr, linewidth=2, c='k', label='degree')

        ax.legend(handles=[degree], fontsize=16)

    else:
        distr = net.get_degree_distr(mode='all')
        outdistr = net.get_degree_distr(mode='out')
        indistr = net.get_degree_distr(mode='in')
        distrlist = [distr, outdistr, indistr]
        if cumulative:
            if survival:
                distrlist = [1 - np.cumsum(d) for d in distrlist]
            else:
                distrlist = [np.cumsum(d) for d in distrlist]

        if log_log:
            degree, = ax.loglog(distrlist[0][:-1], linewidth=2, c='k', label='degree')
            outdegree, = ax.loglog(distrlist[1][:-1], linewidth=2, c='b', label='outdegree')
            indegree, = ax.loglog(distrlist[2][:-1], linewidth=2, c='r', label='indegree')
        else:
            degree, = ax.plot(distrlist[0], linewidth=2, c='k', label='degree')
            outdegree, = ax.plot(distrlist[1], linewidth=2, c='b', label='outdegree')
            indegree, = ax.plot(distrlist[2], linewidth=2, c='r', label='indegree')

        ax.legend(handles=[degree, outdegree, indegree], fontsize=16)
        
def draw_spectrum(net, mode='adj', ax=None, colors=None, cmap='plasma', nbins=None):
    spectrum = net.get_spectrum(mode)
    data = np.array(sorted(list(set(spectrum)), key=np.abs))

    if ax is None:
        fig, ax = create_default_figure(12,10)

    if net.directed:
        ax.scatter(data.real, data.imag, cmap=cmap, c=colors)
    else:
        if nbins is None:
            nbins = len(spectrum) / 10
        ax.hist(data.real, bins=nbins)
        

def get_vector_coloring(vec, cmap='plasma'):
    cmap = plt.get_cmap(cmap)
    vec = np.array(vec).ravel()
    colors = cmap((vec-min(vec))/(max(vec)-min(vec)))
    return colors


def draw_eigenvectors(net, left_ind, right_ind, mode='adj', nodesize=None,
                       cmap='plasma', draw_edges=True, edge_options={}):
    spectrum = net.get_spectrum(mode)
    eigenvectors = net.get_eigenvectors(mode)

    vecs = eigenvectors[:, left_ind: right_ind + 1]
    # vecs = np.abs(net.eigenvectors[:, left_ind: right_ind+1])
    eigvals = np.real(spectrum[left_ind: right_ind + 1])

    npics = vecs.shape[1]
    pics_in_a_row = int(np.ceil(np.sqrt(npics)))
    pics_in_a_col = int(np.ceil(1.0 * npics / pics_in_a_row))
    fig, axs = plt.subplots(nrows=pics_in_a_col, ncols=pics_in_a_row, figsize=(8 * pics_in_a_row, 8 * pics_in_a_col))
    for ax in axs.ravel():
        ax.set_axis_off()
    plt.subplots_adjust(left=0.1, bottom=0.1, right=0.9, top=0.9,
                        hspace=0.2, wspace=0.1)

    if net.pos is None:
        pos = nx.layout.spring_layout(net.graph)
        # pos = nx.drawing.layout.circular_layout(net.graph)
    else:
        pos = net.pos

    if nodesize is None:
        nodesize = np.sqrt(net.scaled_outdeg) * 500 + 50

    for i in range(npics):
        vec = vecs[:, i]
        ax = fig.add_subplot(pics_in_a_col, pics_in_a_row, i + 1)

        text = 'eigenvector ' + str(i + 1) + ' lambda ' + str(np.round(eigvals[i], 3))
        ax.set_title(text)

        ncolors = get_vector_coloring(vec)
        options = {
            'node_color': ncolors,
            'node_size': nodesize,
            'cmap': mpl.colormaps[cmap]
        }

        nodes = nx.draw_networkx_nodes(net.graph, pos, ax=ax, **options)

        if draw_edges:
            edges = nx.draw_networkx_edges(net.graph, pos, **edge_options)
        # pc, = mpl.collections.PatchCollection(nodes, cmap = options['cmap'])
        # pc.set_array(edge_colors)

        cmappable = ScalarMappable(color_normalize(0, 1), cmap=cmap)
        fig.colorbar(cmappable, ax=ax)
        ax.set_axis_off()

    plt.tight_layout()
    plt.show()


def draw_net(net, colors=None, nodesize=None, ax=None):
    if ax is None:
        fig, ax = plt.subplots(figsize=(16, 12))

    if net.pos is None:
        print('Node positions not found, auto layout was constructed')
        pos = nx.layout.spring_layout(net.graph)
        # pos = nx.drawing.layout.circular_layout(net.graph)
    else:
        pos = net.pos

    if nodesize is None:
        nodesize = np.sqrt(net.scaled_outdeg) * 100 + 10

    node_options = {
        'node_size': nodesize,
        'cmap': cm.get_cmap('Spectral'),
    }
    edge_options = {}

    nodes = nx.draw_networkx_nodes(net.graph, pos, node_color=colors, ax=ax, **node_options)
    edges = nx.draw_networkx_edges(net.graph, pos, ax=ax, **edge_options)

    plt.show()
        

def show_mat(net, dtype=None, mode='adj', ax=None):
    mat = getattr(net, mode)
    if mat is None:
        if mode in ['lap', 'lap_out']:
            mat = get_laplacian(net.adj)
        elif mode == 'nlap':
            mat = get_norm_laplacian(net.adj)

    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 10))

    if not dtype is None:
        ax.matshow(mat.astype(dtype).A)
    else:
        ax.matshow(mat.A)
        
        
def plot_lem_embedding(net, ndim, colors=None):

    if net.lem_emb is None:
        net.construct_lem_embedding(ndim)

    if colors is None:
        colors = np.zeros(net.lem_emb.shape[1])
        colors = range(net.lem_emb.shape[1])

    psize = 10
    data = net.lem_emb.A
    pairs = list(combinations(np.arange(ndim), 2))
    npics = len(pairs)
    pics_in_a_row = np.ceil(np.sqrt(npics))
    pics_in_a_col = np.ceil(1.0 * npics / pics_in_a_row)

    fig = plt.figure(figsize=(20, 20))
    fig.suptitle('Projections')
    for i in range(len(pairs)):
        ax = fig.add_subplot(pics_in_a_row, pics_in_a_col, i + 1)
        i1, i2 = pairs[i]
        scatter = ax.scatter(data[i1, :], data[i2, :], c=colors, s=psize)

        legend = ax.legend(*scatter.legend_elements(),
                           loc="upper left", title="Classes")

        ax.text(min(data[i1, :]), min(data[i2, :]), 'axes ' + str(i1 + 1) + ' vs.' + str(i2 + 1),
                bbox={'facecolor': 'red', 'alpha': 0.5, 'pad': 10})
        # ax.add_artist(legend)


import numpy as np
import tqdm

from .stats import *
from ..information.info_base import get_1d_mi, get_multi_mi


def calculate_optimal_delays(ts_bunch1, ts_bunch2, shift_window, ds, verbose=True):
    if verbose:
        print('Calculating optimal delays:')

    optimal_delays = np.zeros((len(ts_bunch1), len(ts_bunch2)), dtype=int)
    shifts = np.arange(-shift_window, shift_window, ds) // ds
    for i, ts1 in tqdm.tqdm(enumerate(ts_bunch1), total=len(ts_bunch1)):
        for j, ts2 in enumerate(ts_bunch2):
            shifted_mi = []
            for shift in shifts:
                lag_mi = get_1d_mi(ts1, ts2, ds=ds, shift=int(shift))
                shifted_mi.append(lag_mi)

            best_shift = shifts[np.argmax(shifted_mi)]
            optimal_delays[i, j] = int(best_shift*ds)

    return optimal_delays


# TODO: add cbunch and fbunch logic
def get_calcium_feature_mi_profile(exp, cell_id, feat_id, window=1000, ds=1):

    cell = exp.neurons[cell_id]
    ts1 = cell.ca
    shifted_mi = []

    if isinstance(feat_id, str):
        ts2 = exp.dynamic_features[feat_id]
        mi0 = get_1d_mi(ts1, ts2, ds=ds)

        for shift in tqdm.tqdm(np.arange(-window, window, ds)//ds):
            lag_mi = get_1d_mi(ts1, ts2, ds=ds, shift=shift)
            shifted_mi.append(lag_mi)

    else:
        feats = [exp.dynamic_features[fid] for fid in feat_id]
        mi0 = get_multi_mi(feats, ts1, ds=ds)

        for shift in tqdm.tqdm(np.arange(-window, window, ds)):
            lag_mi = get_multi_mi(feats, ts1, ds=ds, shift=shift)
            shifted_mi.append(lag_mi)

    return mi0, shifted_mi


def scan_pairs(ts_bunch1,
               ts_bunch2,
               nsh,
               optimal_delays,
               joint_distr=False,
               ds=1,
               mask=None,
               noise_const=1e-3,
               seed=None):

    """
    Calculates MI shuffles for 2 given sets of TimeSeries
    This function is generally assumed to be used internally,
    but can be also called manually to "look inside" high-level computation routines

    Parameters
    ----------
    ts_bunch1: list of TimeSeries objects

    ts_bunch2: list of TimeSeries objects

    nsh: int
        number of shuffles

    joint_distr: bool
        if joint_distr=True, ALL (sic!) TimeSeries in ts_bunch2 will be treated as components of a single multifeature
        default: False

    ds: int
        Downsampling constant. Every "ds" point will be taken from the data time series.
        default: 1

    mask: np.array of shape (len(ts_bunch1), len(ts_bunch2)) or (len(ts_bunch), 1) if joint_distr=True
          precomputed mask for skipping some of possible pairs.
          0 in mask values means calculation will be skipped.
          1 in mask values means calculation will proceed.

    noise_const: float
        Small noise amplitude, which is added to MI and shuffled MI to improve numerical fit
        default: 1e-3

    optimal_delays: np.array of shape (len(ts_bunch1), len(ts_bunch2)) or (len(ts_bunch), 1) if joint_distr=True
        best shifts from original time series alignment in terms of MI.

    seed: int
        Random seed for reproducibility

    Returns
    -------
    random_shifts: np.array of shape (nsh, len(cells))
        signals shifts used for MI distribution computation

    mi_total: np.array of shape (len(ts_bunch1), len(ts_bunch2)), nsh+1) or (len(ts_bunch1), 1, nsh+1) if joint_distr=True
        Aggregated array of true and shuffled MI values.
        True MI matrix can be obtained by mi_total[:,:,0]
        Shuffled MI tensor of shape (len(ts_bunch1), len(ts_bunch2)), nsh) or (len(ts_bunch1), 1, nsh) if joint_distr=True
        can be obtained by mi_total[:,:,1:]
    """

    if seed is not None:
        np.random.seed = seed

    n1 = len(ts_bunch1)
    n2 = 1 if joint_distr else len(ts_bunch2)
    t = len(ts_bunch1[0].data)  # full length is the same for all time series

    if mask is None:
        mask = np.ones((n1, n2))

    mi_table = np.zeros((n1, n2))
    mi_table_shuffles = np.zeros((n1, n2, nsh))
    random_shifts = np.zeros((n1, n2, nsh), dtype=int)

    # fill random shifts according to the allowed shuffles masks of both time series
    for i, ts1 in enumerate(ts_bunch1):
        if joint_distr:
            # TODO: add combination of ts shuffle masks for all ts from tsbunch2
            combined_shuffle_mask = ts1.shuffle_mask
            # move shuffle mask according to optimal shift
            combined_shuffle_mask = np.roll(combined_shuffle_mask, int(optimal_delays[i, 0]))
            indices_to_select = np.arange(t)[combined_shuffle_mask]
            random_shifts[i, 0, :] = np.random.choice(indices_to_select, size=nsh) // ds

        else:
            for j, ts2 in enumerate(ts_bunch2):
                combined_shuffle_mask = ts1.shuffle_mask & ts2.shuffle_mask
                # move shuffle mask according to optimal shift
                combined_shuffle_mask = np.roll(combined_shuffle_mask, int(optimal_delays[i, j]))
                indices_to_select = np.arange(t)[combined_shuffle_mask]
                random_shifts[i, j, :] = np.random.choice(indices_to_select, size=nsh)//ds

    for i, ts1 in tqdm.tqdm(enumerate(ts_bunch1), total=len(ts_bunch1), position=0, leave=True):
        if joint_distr:
            if mask[i,0] == 1:
                # default MI without shuffling, minus due to different order
                mi0 = get_multi_mi(ts_bunch2, ts1, ds=ds, shift=-optimal_delays[i, 0]//ds)
                mi_table[i,0] = mi0 + np.random.random()*noise_const  # add small noise for better fitting

                for k, shift in enumerate(random_shifts[i, 0, :]):
                    mi = get_multi_mi(ts_bunch2, ts1, ds=ds, shift=shift)
                    mi_table_shuffles[i,0,k] = mi + np.random.random()*noise_const  # add small noise for better fitting

            else:
                mi_table[i,0] = None
                mi_table_shuffles[i,0,:] = np.full(shape=nsh, fill_value=None)

        else:
            for j, ts2 in enumerate(ts_bunch2):
                if mask[i,j] == 1:
                    mi0 = get_1d_mi(ts1, ts2, ds=ds, shift=optimal_delays[i,j]//ds) # default MI without shuffling
                    mi_table[i,j] = mi0 + np.random.random()*noise_const  # add small noise for better fitting

                    for k, shift in enumerate(random_shifts[i,j,:]):
                        mi = get_1d_mi(ts1, ts2, shift=shift, ds=ds)
                        mi_table_shuffles[i,j,k] = mi + np.random.random()*noise_const  # add small noise for better fitting

                else:
                    mi_table[i,j] = None
                    mi_table_shuffles[i,j,:] = np.array([None for _ in range(nsh)])

    mi_total = np.dstack((mi_table, mi_table_shuffles))

    return random_shifts, mi_total


def compute_mi_stats(ts_bunch1,
                     ts_bunch2,
                     names1=None,
                     names2=None,
                     mode='two_stage',
                     precomputed_mask_stage1=None,
                     precomputed_mask_stage2=None,
                     n_shuffles_stage1=100,
                     n_shuffles_stage2=10000,
                     joint_distr=False,
                     mi_distr_type='gamma',
                     noise_ampl=1e-3,
                     ds=1,
                     topk1=1,
                     topk2=5,
                     multicomp_correction='holm',
                     pval_thr=0.01,
                     find_optimal_delays=False,
                     shift_window=100,
                     verbose=True,
                     seed=None):

    """
    Calculates MI statistics for TimeSeries pairs

    Parameters
    ----------
    ts_bunch1: list of TimeSeries objects

    ts_bunch2: list of TimeSeries objects

    names1: list of str
        names than will be given to time series from tsbunch1 in final results

    names2: list of str
        names than will be given to time series from tsbunch2 in final results

    mode: str
        Computation mode. 3 modes are available:
        'stage1': perform preliminary scanning with "n_shuffles_stage1" shuffles only.
                  Rejects strictly non-significant neuron-feature pairs, does not give definite results
                  about significance of the others.
        'stage2': skip stage 1 and perform full-scale scanning ("n_shuffles_stage2" shuffles) of all neuron-feature pairs.
                  Gives definite results, but can be very time-consuming. Also reduces statistical power
                  of multiple comparison tests, since the number of hypotheses is very high.
        'two_stage': prune non-significant pairs during stage 1 and perform thorough testing for the rest during stage 2.
                     Recommended mode.
        default: 'two-stage'

    precomputed_mask_stage1: np.array of shape (len(ts_bunch1), len(ts_bunch2)) or (len(ts_bunch), 1) if joint_distr=True
          precomputed mask for skipping some of possible pairs in stage 1.
          0 in mask values means calculation will be skipped.
          1 in mask values means calculation will proceed.

    precomputed_mask_stage2: np.array of shape (len(ts_bunch1), len(ts_bunch2)) or (len(ts_bunch), 1) if joint_distr=True
          precomputed mask for skipping some of possible pairs in stage 2.
          0 in mask values means calculation will be skipped.
          1 in mask values means calculation will proceed.

    n_shuffles_stage1: int
        number of shuffles for first stage
        default: 100

    n_shuffles_stage2: int
        number of shuffles for second stage
        default: 10000

    joint_distr: bool
        if joint_distr=True, ALL features in feat_bunch will be treated as components of a single multifeature
        For example, 'x' and 'y' features will be put together into ('x','y') multifeature.
        default: False

    mi_distr_type: str
        Distribution type for shuffles MI distribution fit. Supported options are "gamma" and "lognormal"
        default: "gamma"

    noise_ampl: float
        Small noise amplitude, which is added to MI and shuffled MI to improve numerical fit
        default: 1e-3

    ds: int
        Downsampling constant. Every "ds" point will be taken from the data time series.
        default: 1

    topk1: int
        true MI for stage 1 should be among topk1 MI shuffles
        default: 1

    topk2: int
        true MI for stage 2 should be among topk2 MI shuffles
        default: 5

    multicomp_correction: str or None
        type of multiple comparisons correction. Supported types are None (no correction),
        "bonferroni" and "holm".
        default: 'holm'

    pval_thr: float
        pvalue threshold. if multicomp_correction=None, this is a p-value for a single pair.
        Otherwise it is a FWER significance level.

    find_optimal_delays: bool
        Allows slight shifting (not more than +- shift_window) of time series,
        selects a shift with the highest MI as default.
        default: True

    shift_window: int
        Window for optimal shift search (frames). Optimal shift will lie in the range
        -shift_window <= opt_shift <= shift_window

    verbose: bool
        whether to print intermediate information

    seed: int
        random seed for reproducibility

    Returns
    -------
    stats: dict of dict of dicts
        Outer dict keys: indices of tsbunch1 or names1, if given
        Inner dict keys: indices or tsbunch2 or names2, if given
        Last dict: dictionary of stats variables.
        Can be easily converted to pandas DataFrame by pd.DataFrame(stats)

    significance: dict of dict of dicts
        Outer dict keys: indices of tsbunch1 or names1, if given
        Inner dict keys: indices or tsbunch2 or names2, if given
        Last dict: dictionary of significance-related variables.
        Can be easily converted to pandas DataFrame by pd.DataFrame(significance)

    accumulated_info: dict
        Data collected during computation.
    """

    # TODO: add automatic min_shifts from autocorrelation time

    accumulated_info = dict()

    n1 = len(ts_bunch1)
    n2 = 1 if joint_distr else len(ts_bunch2)

    if precomputed_mask_stage1 is None:
        precomputed_mask_stage1 = np.ones((n1, n2))
    if precomputed_mask_stage2 is None:
        precomputed_mask_stage2 = np.ones((n1, n2))

    optimal_delays = np.zeros((n1, n2))
    if find_optimal_delays:
        optimal_delays = calculate_optimal_delays(ts_bunch1, ts_bunch2, shift_window, ds, verbose=verbose)
    accumulated_info['optimal_delays'] = optimal_delays

    mask_from_stage1 = np.zeros((n1, n2))
    mask_from_stage2 = np.zeros((n1, n2))
    nhyp = n1*n2

    if mode in ['two_stage', 'stage1']:
        npairs_to_check1 = int(np.sum(precomputed_mask_stage1))
        if verbose:
            print(f'Starting stage 1 scanning for {npairs_to_check1}/{nhyp} possible pairs')

        # STAGE 1 - primary scanning
        random_shifts1, mi_total1 = scan_pairs(ts_bunch1,
                                               ts_bunch2,
                                               n_shuffles_stage1,
                                               optimal_delays,
                                               joint_distr=joint_distr,
                                               ds=ds,
                                               mask=precomputed_mask_stage1,
                                               noise_const=noise_ampl,
                                               seed=seed)

        accumulated_info.update({'random_shifts1': random_shifts1,
                                'mi_total1': mi_total1})
        # turn computed data tables from stage 1 and precomputed data into dict of stats dicts
        stage_1_stats = get_table_of_stats(mi_total1,
                                           optimal_delays,
                                           mi_distr_type=mi_distr_type,
                                           nsh=n_shuffles_stage1,
                                           precomputed_mask=precomputed_mask_stage1,
                                           stage=1)

        # select potentially significant pairs for stage 2
        # 0 in mask values means the pair MI is definitely insignificant, stage 2 calculation will be skipped.
        # 1 in mask values means the pair MI is potentially significant, stage 2 calculation will proceed.

        if verbose:
            print('Computing significance for all pairs in stage 1...')

        stage_1_significance = populate_nested_dict(dict(), range(n1), range(n2))
        for i in range(n1):
            for j in range(n2):
                pair_passes_stage1 = criterion1(stage_1_stats[i][j],
                                                n_shuffles_stage1,
                                                topk=topk1)
                if pair_passes_stage1:
                    mask_from_stage1[i, j] = 1

                sig1 = {'shuffles1': n_shuffles_stage1,
                        'stage1': pair_passes_stage1}
                stage_1_significance[i][j].update(sig1)

        nhyp = int(np.sum(mask_from_stage1))  # number of hypotheses for further statistical testing
        if verbose:
            print('Stage 1 results:')
            print(f'{nhyp/n1/n2*100:.2f}% ({nhyp}/{n1*n2}) of possible pairs identified as candidates')

    if mode == 'stage1' or nhyp == 0:
        final_stats = add_names_to_nested_dict(stage_1_stats, names1, names2)
        final_significance = add_names_to_nested_dict(stage_1_significance, names1, names2)

        return final_stats, final_significance, accumulated_info

    else:
        # STAGE 2 - full-scale scanning
        combined_mask_for_stage_2 = np.ones((n1, n2))
        combined_mask_for_stage_2[np.where(mask_from_stage1 == 0)] = 0  # exclude non-significant pairs from stage1
        combined_mask_for_stage_2[np.where(precomputed_mask_stage2 == 0)] = 0  # exclude precomputed stage 2 pairs

        npairs_to_check2 = int(np.sum(combined_mask_for_stage_2))
        if verbose:
            print(f'Starting stage 2 scanning for {npairs_to_check2}/{nhyp} possible pairs')

        random_shifts2, mi_total2 = scan_pairs(ts_bunch1,
                                               ts_bunch2,
                                               n_shuffles_stage2,
                                               optimal_delays,
                                               joint_distr=joint_distr,
                                               ds=ds,
                                               mask=combined_mask_for_stage_2,
                                               noise_const=noise_ampl,
                                               seed=seed)

        accumulated_info.update({'random_shifts2': random_shifts2,
                                 'mi_total2': mi_total2})

        # turn data tables from stage 2 to array of stats dicts
        stage_2_stats = get_table_of_stats(mi_total2,
                                           optimal_delays,
                                           mi_distr_type=mi_distr_type,
                                           nsh=n_shuffles_stage2,
                                           precomputed_mask=combined_mask_for_stage_2,
                                           stage=2)

        # select significant pairs after stage 2
        if verbose:
            print('Computing significance for all pairs in stage 2...')
        all_pvals = None
        if multicomp_correction == 'holm':  # holm procedure requires all p-values
            all_pvals = get_all_nonempty_pvals(stage_2_stats, range(n1), range(n2))

        multicorr_thr = get_multicomp_correction_thr(pval_thr,
                                                     mode=multicomp_correction,
                                                     all_pvals=all_pvals,
                                                     nhyp=nhyp)

        stage_2_significance = populate_nested_dict(dict(), range(n1), range(n2))
        for i in range(n1):
            for j in range(n2):
                pair_passes_stage2 = criterion2(stage_2_stats[i][j],
                                                n_shuffles_stage2,
                                                multicorr_thr,
                                                topk=topk2)
                if pair_passes_stage2:
                    mask_from_stage2[i,j] = 1

                sig2 = {'shuffles2': n_shuffles_stage2,
                        'stage2': pair_passes_stage2,
                        'final_p_thr': pval_thr,
                        'multicomp_corr': multicomp_correction,
                        'pairwise_pval_thr': multicorr_thr}

                stage_2_significance[i][j] = sig2

        num2 = int(np.sum(mask_from_stage2))
        if verbose:
            print('Stage 2 results:')
            print(f'{num2/n1/n2*100:.2f}% ({num2}/{n1*n2}) of possible pairs identified as significant')

        if mode == 'two_stage':
            merged_stats = merge_stage_stats(stage_1_stats, stage_2_stats)
            merged_significance = merge_stage_significance(stage_1_significance, stage_2_significance)
            final_stats = add_names_to_nested_dict(merged_stats, names1, names2)
            final_significance = add_names_to_nested_dict(merged_significance, names1, names2)
            return final_stats, final_significance, accumulated_info

        else:
            return stage_2_stats, stage_2_significance, accumulated_info


def get_multicomp_correction_thr(fwer, mode='holm', **multicomp_kwargs):

    '''
    Calculates pvalue threshold for a single hypothesis from FWER

    Parameters
    ----------
    fwer: float
        family-wise error rate

    mode: str or None
        type of multiple comparisons correction. Supported types are None (no correction),
        "bonferroni" and "holm".

    multicomp_kwargs: named arguments for multiple comparisons correction procedure
    '''

    if mode is None:
        threshold = fwer

    elif mode == 'bonferroni':
        if 'nhyp' in multicomp_kwargs:
            threshold = fwer / multicomp_kwargs['nhyp']
        else:
            raise ValueError('Number of hypotheses for Bonferroni correction not provided')

    elif mode == 'holm':
        if 'all_pvals' in multicomp_kwargs:
            all_pvals = sorted(multicomp_kwargs['all_pvals'])
            nhyp = len(all_pvals)
            for i, pval in enumerate(all_pvals):
                cthr = fwer / (nhyp - i)
                if pval > cthr:
                    break

            threshold = cthr
        else:
            raise ValueError('List of p-values for Holm correction not provided')

    else:
        raise ValueError('Unknown multiple comparisons correction method')

    return threshold



import numpy as np
import copy
import scipy.sparse as sp


def _plain_bfs(adj, source):
    '''
    adapted from networkx.algorithms.components.connected._plain_bfs

    Args:
        adj:
        source:

    Returns:

    '''

    n = adj.shape[0]
    seen = {source}
    nextlevel = [source]
    while nextlevel:
        thislevel = nextlevel
        nextlevel = []
        for v in thislevel:
            for w in get_neighbors_from_adj(adj, v):
                if w not in seen:
                    seen.add(w)
                    nextlevel.append(w)
            if len(seen) == n:
                return seen
    return seen


def get_neighbors_from_adj(a, node):
    inds = a[[node], :].nonzero()[1]
    return inds


def get_ccs_from_adj(adj):
    seen = set()
    for v in range(adj.shape[0]):
        if v not in seen:
            c = _plain_bfs(adj, v)
            seen.update(c)
            yield c


def get_sccs_from_adj(adj):
    '''
        adapted from networkx.algorithms.components.strongly_connected.strongly_connected_components
    Args:
        adj:

    Returns:

    '''

    all_nodes = range(adj.shape[0])
    preorder = {}
    lowlink = {}
    scc_found = set()
    scc_queue = []
    i = 0  # Preorder counter
    neighbors = {v: iter(get_neighbors_from_adj(adj, v)) for v in all_nodes}
    for source in all_nodes:
        if source not in scc_found:
            queue = [source]
            while queue:
                v = queue[-1]
                if v not in preorder:
                    i = i + 1
                    preorder[v] = i
                done = True
                for w in neighbors[v]:
                    if w not in preorder:
                        queue.append(w)
                        done = False
                        break
                if done:
                    lowlink[v] = preorder[v]
                    for w in get_neighbors_from_adj(adj, v):
                        if w not in scc_found:
                            if preorder[w] > preorder[v]:
                                lowlink[v] = min([lowlink[v], lowlink[w]])
                            else:
                                lowlink[v] = min([lowlink[v], preorder[w]])
                    queue.pop()
                    if lowlink[v] == preorder[v]:
                        scc = {v}
                        while scc_queue and preorder[scc_queue[-1]] > preorder[v]:
                            k = scc_queue.pop()
                            scc.add(k)
                        scc_found.update(scc)
                        yield scc
                    else:
                        scc_queue.append(v)


def get_giant_cc_from_adj(adj):
    connected_components = sorted(get_ccs_from_adj(adj), key=len, reverse=True)
    gcc = np.array(list(connected_components[0]))
    gcc_adj = adj[gcc, :].tocsc()[:, gcc].tocsr()

    # mapping of new nodes to old ones
    node_mapping = dict(zip(range(len(gcc)), gcc))

    return gcc_adj, node_mapping


def get_giant_scc_from_adj(adj):
    connected_components = sorted(get_sccs_from_adj(adj), key=len, reverse=True)
    gscc = np.array(list(connected_components[0]))
    gscc_adj = adj[gscc, :].tocsc()[:, gscc].tocsr()

    # mapping of new nodes to old ones
    node_mapping = dict(zip(range(len(gscc)), gscc))

    return gscc_adj, node_mapping


def assign_random_weights(A):
    X = np.random.random(size=(A.shape[0], A.shape[0]))
    W = np.multiply(X, A)
    return (W + W.T) / 2


# TODO: refactor to sparse format
def turn_to_partially_directed(mat, directed=0.0, weighted=0):
    if not isinstance(mat, np.ndarray):
        raise Exception('Wrong input parsed to turn_to_directed function!')

    A = copy.deepcopy(mat)

    if directed == 0.0 or directed is None:
        if not weighted:
            a = A.astype(bool)
        else:
            a = A.astype(float)

        return sp.csr_array(a)

    np.fill_diagonal(A, 0)
    rows, cols = A.nonzero()
    edgeset = set(zip(rows, cols))
    upper = np.array([l for l in edgeset if l[0] < l[1]])
    dircount = 0

    random_tosses = np.random.random(len(upper))
    condition1 = (random_tosses >= directed / 2.0) & (random_tosses < directed)
    condition2 = (random_tosses <= directed / 2.0) & (random_tosses < directed)
    indices_where_upper_is_removed = np.where(condition1 == True)[0]
    indices_where_lower_is_removed = np.where(condition2 == True)[0]

    u_xdata = [u[0] for u in upper[indices_where_upper_is_removed]]
    u_ydata = [u[1] for u in upper[indices_where_upper_is_removed]]
    A[u_xdata, u_ydata] = 0

    l_xdata = [u[1] for u in upper[indices_where_lower_is_removed]]
    l_ydata = [u[0] for u in upper[indices_where_lower_is_removed]]
    A[l_xdata, l_ydata] = 0

    '''
    for i in range(double_edges):
        toss = random.random()
        if toss < directed: # this means double edge will be reduced to single randomly
            dircount += 1
            if toss >= directed/2.:
                A[upper_right[i]] = 0#A[upper_right[i][::-1]] + 0#.1*np.random.random()
            else:
                A[upper_right[i][::-1]] = 0#A[upper_right[i]] + 0#.1*np.random.random()
    '''

    #a = sp.csr_array(A)
    # get_symmetry_index(a)
    return A


def get_symmetry_index(a):
    a = a.astype(bool)
    symmetrized = a + a.T

    difference = symmetrized.astype(int) - a.astype(int)
    difference.eliminate_zeros()
    symm_index = 1 - difference.nnz / symmetrized.nnz * 2
    # symm_index is 1 for a symmetrix matrix and 0 for an asymmetric one
    return symm_index


def symmetric_component(A, is_weighted):
    a = A.astype(bool).A
    symm_mask = np.bitwise_and(a, a.T)
    if not is_weighted:
        return symm_mask

    return np.multiply(symm_mask, A.A)


def non_symmetric_component(A, is_weighted):
    return A.astype(float) - symmetric_component(A, is_weighted).astype(float)


def remove_duplicates(coo):
    # this function removes duplicate entries from a coo-format adjacency matrix
    # duplicates are discarded as the data is always the same:
    # coo[i,j] = val1, coo[i,j] = val2 ==> val1 = val2

    dok = sp.dok_matrix((coo.shape), dtype=coo.dtype)
    dok._update(zip(zip(coo.row, coo.col), coo.data))
    return dok.tocoo()


def adj_input_to_csr_sparse_matrix(a):
    if isinstance(a, np.ndarray):
        adj = sp.csr_array(a)
    elif a.format in ['csr', 'csc']:
        adj = a
    elif a.format == 'coo':
        adj = remove_duplicates(a)
    else:
        raise Exception('Wrong input parsed to preprocess_adj_matrix function:', type(a))

    return sp.csr_array(adj)


def remove_selfloops_from_adj(a):
    if a.trace() != 0:
        a = adj_input_to_csr_sparse_matrix(a)
        anew = a.copy()
        anew.setdiag(0)
        anew.eliminate_zeros()
        return anew
    else:
        return a


def remove_isolates_from_adj(a):
    a = adj_input_to_csr_sparse_matrix(a)

    in_degrees = np.array(a.astype(bool).astype(int).sum(axis=1))  # .flatten().ravel()
    out_degrees = np.array(a.astype(bool).astype(int).sum(axis=0))  # .flatten().ravel()

    indices = np.where(in_degrees + out_degrees > 0)[0]
    cleared_matrix = a[indices, :].tocsc()[:, indices].tocsr()

    # mapping of new nodes to old ones
    node_mapping = dict(zip(range(len(indices)), indices))

    return cleared_matrix, node_mapping


def sausage_index(A, nn):
    A = A.astype(bool).astype(int)
    sausage_edges = 0
    for i in range(nn):
        sausage_edges += sum(np.diag(A, k=i))

    si = sausage_edges / (np.sum(A) / 2)
    print('sausage edges:', sausage_edges)
    print('other edges:', np.sum(A) / 2 - sausage_edges)
    print('sausage index=', si)


#TODO: create separate branches for np and sparse matrices for further functions
def get_laplacian(A):
    A = A.astype(float)
    out_degrees = np.array(A.sum(axis=0)).ravel()
    D = sp.spdiags(out_degrees, [0], A.shape[0], A.shape[0], format='csr')
    L = D - A
    return L


def get_inv_sqrt_diag_matrix(a):
    n = a.shape[0]
    A = sp.csr_array(a)
    out_degrees = np.array(A.sum(axis=0)).ravel()
    diags_sqrt = 1.0 / np.sqrt(out_degrees)

    diags_sqrt[np.isinf(diags_sqrt)] = 0
    DH = sp.spdiags(diags_sqrt, [0], n, n, format='csr')
    return DH


def get_norm_laplacian(a):
    if get_symmetry_index(a) != 1:
        raise Exception('Cannot construct normalized laplacian matrix from a non-hermitian adjacency matrix')

    n = a.shape[0]
    A = sp.csr_array(a)
    DH = get_inv_sqrt_diag_matrix(A)
    matrix = sp.eye(n, dtype=float) - DH.dot(A.dot(DH))
    return matrix


def get_inv_diag_matrix(a):
    n = a.shape[0]
    A = sp.csr_array(a)
    out_degrees = np.array(A.sum(axis=0)).ravel()
    invdiags = 1.0 / out_degrees

    invdiags[np.isinf(invdiags)] = 0
    Dinv = sp.spdiags(invdiags, [0], n, n, format='csr')
    return Dinv


def get_rw_laplacian(a):
    n = a.shape[0]
    T = get_trans_matrix(a)
    matrix = sp.eye(n, dtype=float) - T
    return matrix


def get_trans_matrix(a):
    A = sp.csr_array(a)
    Dinv = get_inv_diag_matrix(a)
    T = Dinv.dot(A)
    return T



import numpy as np


def free_entropy(spectrum, t):
    # eq.2 in https://www.nature.com/articles/s42005-021-00582-8#Sec10
    eigenvalues = np.exp(-t * spectrum)
    F = np.log2(np.real(np.sum(eigenvalues)))
    return F


def q_entropy(spectrum, t, q=1):
    """

    Args:
        spectrum:
        t:
        q:

    Returns:

    """
    # https://journals.aps.org/prx/abstract/10.1103/PhysRevX.6.041062

    if q <= 0:
        raise Exception('q must be >0')
    else:
        Z = np.sum(np.exp(-t * spectrum))
        if q != 1:
            eigenvalues = np.exp(-t * q * spectrum)
            S = 1/(1-q) * np.log(Z**(-q) * np.sum(eigenvalues))
        else:
            S = spectral_entropy(spectrum, t, verbose=0)

    if np.imag(S) != 0:
        raise Exception(f'Imaginary entropy detected: t={t}, q={q}, S={S}!')

    return S


def spectral_entropy(spectrum, t, verbose=0):
    eigenvalues = np.exp(-t * spectrum)
    norm_eigenvalues = np.trim_zeros(eigenvalues / np.sum(eigenvalues))
    S = -np.real(np.sum(np.multiply(norm_eigenvalues, np.log2(norm_eigenvalues))))

    if verbose:
        print('initial eigenvalues:')
        print(spectrum)
        print('exp eigenvalues:')
        print(eigenvalues)
        print('norm exp eigenvalues:')
        print(norm_eigenvalues)
        print('logs:')
        print(np.log2(norm_eigenvalues))

    return S


from scipy import signal
import numpy as np
from numpy.fft import rfft, irfft
import pywt
from .complexity import ApEn
from ..information.info_base import TimeSeries

def multiopt(x, thr):
    if thr == 0.0:
        return x

    m = np.abs(x)

    if m < thr:
        return 0

    return np.sign(x) * (m - 1.0 * thr * np.exp(1 - 1.0 * m / thr))


def truncate_sig(coeffs, thr):
    return np.array([multiopt(x, thr) for x in coeffs])


def ts_wavelet_denoise(d, denoising_params):
    sig = Signal(d)
    sig.get_idwt(**denoising_params)
    try:
        new_d = sig.idwt[:-1]
    except:
        new_d = sig.idwt

    return new_d


# TODO: refactor Signal class into TimeSeries child class
class Signal(TimeSeries):
    '''
    Class implementing additional functionality for continuous time series
    '''
    def __init__(self, data):
        super().__init__(data, discrete=False)

        self.ft = None
        self.cwt = None
        self.dwt = None
        self.hurst_exp = None

        self.mod_ft = None
        self.mod_cwt = None
        self.mod_dwt = None

        self.ift = None
        self.icwt = None
        self.idwt = None

        self.ap_en = None

    def get(self, fn, args=None):
        if not hasattr(self, fn):
            raise Exception('Wrong signals property!')

        d = getattr(self, fn)
        if d is None:
            constructor = 'get_' + fn
            getattr(self, constructor)(args)
            d = getattr(self, fn)

        return d

    def get_ft(self):
        sig = self.data
        ft = rfft(sig)
        self.ft = ft

    def get_ift(self, thr):
        ft = self.get('ft')
        mod_ft = ft.copy()
        # print(mod_ft.shape)
        mod_ft[(mod_ft > thr)] = 0
        self.ift = irfft(mod_ft)

    def modify_dwt(self, cd, thr):
        cd_mod = truncate_sig(cd, thr)
        return cd_mod

    def get_cwt(self, max_width=16):
        sig = self.data
        wavelet = signal.ricker
        # wavelet = signals.morlet
        widths = np.arange(1, max_width)
        cwt_mat = signal.cwt(sig, wavelet, widths)
        self.cwt = cwt_mat

    def get_dwt(self, wavelet):
        ca, cd = pywt.dwt(self.data, wavelet)
        self.dwt = (ca, cd)

    def get_idwt(self, wavelet, thr, level):
        n = len(self.data)
        max_level = pywt.dwt_max_level(len(self.data), wavelet)
        if not level is None and level > max_level:
            print('wrong level value, discarded to %s' % max_level)
            level = max_level

        all_coeffs = pywt.wavedec(self.data, wavelet, mode='smooth', level=level)
        # print(sum(all_coeffs[1]))
        for i in range(1, len(all_coeffs)):
            cd = all_coeffs[i]
            sigma = np.median(np.abs(cd)) / 0.6745
            thr = sigma * np.sqrt(2 * np.log(n))
            all_coeffs[i] = self.modify_dwt(cd, thr)

        # restored = pywt.idwt(all_coeffs[0], all_coeffs[1], wavelet, mode = 'smooth')
        # print(sum(all_coeffs[1]))
        restored = pywt.waverec(all_coeffs, wavelet, mode='smooth')
        # print(restored)
        self.idwt = restored

    def get_hurst_exp(self, wavelet, lvl_min, lvl_max):
        n = len(self.data)
        max_possible_level = pywt.dwt_max_level(len(self.data), wavelet)
        if not (lvl_min is None or lvl_max is None) and lvl_max > max_possible_level:
            print('wrong level value, discarded to %s' % max_possible_level)
            lvl_max = max_possible_level

        all_coeffs = pywt.wavedec(self.data, wavelet, mode='smooth', level=lvl_max)
        energies = []
        for i in range(1, len(all_coeffs)):
            if lvl_min <= i and lvl_max >= i:
                dlist = list(all_coeffs[i])
                energies.append(sum([d ** 2 for d in dlist]) / len(dlist))

        x = range(len(energies))[::-1]
        y = np.log2(energies)
        beta = np.polyfit(x, y, 1)[0]
        print('spectrum power = ', beta)
        H = (beta - 1) / 2
        return H

    def ap_en(self):
        if not self.ap_en is None:
            return self.ap_en

        return ApEn(self.data, 3, 1)